####################################################################################
# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".

.
"""
__docformat__ = 'restructuredtext'

import os
import sys
import signal
import uuid
import time

import ansar.create as ar
import ansar.create.point as ap
from .socketry import *
from .directory_if import *
from .directory import *
from .wan import *
from .transporting import *

__all__ = [
	'NodeSettings',
	'node_settings',
	'NodeProperties',
	'node_passing',
	'sub_node_passing',
	'create_node',
]

#
#
class NodeSettings(object):
	def __init__(self, node_scope=None, group_port=None, accept_port=None):
		self.node_scope = node_scope
		self.group_port = group_port
		self.accept_port = accept_port

NODE_SETTINGS_SCHEMA = {
	'node_scope': ScopeOfService,
	'group_port': ar.Integer8(),
	'accept_port': ar.Integer8(),
}

ar.bind(NodeSettings, object_schema=NODE_SETTINGS_SCHEMA)

#
#
node_settings = NodeSettings()


#
#
class NodeProperties(object):
	def __init__(self, guid=None, created=None, executable=None, start_stop=None, retry=None, storage=None):
		self.guid = guid
		self.created = created
		self.executable = executable
		self.start_stop = start_stop or ar.default_deque()
		self.retry = retry or ar.RetryIntervals(step_limit=0)
		self.storage = storage
		#self.connect_above = connect_above					# HostPort or cloud connection or None.
		#self.accept_below = accept_below or HostPort()		# Must be a listen address.

NODE_PROPERTIES_SCHEMA = {
	'guid': ar.UUID(),
	'created': ar.WorldTime(),
	'executable': ar.Unicode(),
	'start_stop': ar.DequeOf(ar.StartStop),
	'retry': ar.UserDefined(ar.RetryIntervals),
	'storage': ar.Integer8(),
}

ar.bind(NodeProperties, object_schema=NODE_PROPERTIES_SCHEMA)

#
#
# Standard parameter processing. Check for name collision.
#
def node_passing(special_settings):
	if special_settings is not None:
		a = node_settings.__art__.value.keys()
		b = ar.object_settings.__art__.value.keys()
		c = special_settings.__art__.value.keys()
		d = set(a) & set(c)
		if len(d) > 0:
			j = ', '.join(c)
			raise ValueError('collision in settings names - {collisions}'.format(collisions=j))
		d = set(b) & set(c)
		if len(d) > 0:
			j = ', '.join(c)
			raise ValueError('collision in settings names - {collisions}'.format(collisions=j))
	executable, word, ls = ar.break_args()
	x1, r1 = ar.extract_args(node_settings, ls, ar.object_settings)
	x2, r2 = ar.extract_args(ar.object_settings, r1, special_settings)
	ar.arg_values(node_settings, x1)
	ar.arg_values(ar.object_settings, x2)
	return executable, word, r2

def sub_node_passing(special_settings, table):
	if special_settings is not None:
		a = node_settings.__art__.value.keys()
		b = ar.object_settings.__art__.value.keys()
		c = special_settings.__art__.value.keys()
		d = set(a) & set(c)
		if len(d) > 0:
			j = ', '.join(c)
			raise ValueError('collision in settings names - {collisions}'.format(collisions=j))
		d = set(b) & set(c)
		if len(d) > 0:
			j = ', '.join(c)
			raise ValueError('collision in settings names - {collisions}'.format(collisions=j))

	executable, ls1, sub, ls2, word = ar.sub_args()
	x1, r1 = ar.extract_args(node_settings, ls1, ar.object_settings)
	x2, r2 = ar.extract_args(ar.object_settings, r1, special_settings)
	ar.arg_values(node_settings, x1)
	ar.arg_values(ar.object_settings, x2)

	# Support for the concept of a noop pass, just for the
	# framework.
	def no_sub_required(s):
		return s.help or s.dump_settings or s.dump_input

	if sub is not None:
		try:
			sub_function, sub_settings = table[sub]
		except KeyError:
			raise ValueError(f'unknown sub-command "{sub}"')

		if sub_settings:
			x3, r3 = ar.extract_args(sub_settings, ls2, None)
			ar.arg_values(sub_settings, x3)
		else:
			r3 = ls2
	elif no_sub_required(ar.object_settings):
		# Give framework a chance to complete some
		# admin operation.
		sub_function = None
		r3 = ({}, {})
	else:
		raise ValueError('no-op command')

	bundle = (sub_function, # The sub-command function.
		r3,				 # Remainder from ls2, i.e. for passing to sub-component
		word)			   # Non-flag arguments.

	return executable, bundle, r2

#
#
def node_vector(self, object_type, settings, input, variables, fixed_value, key_value):
	role = ar.object_role()

	name_counts = ['"%s" (%d)' % (k, len(v)) for k, v in ap.pt.thread_classes.items()]

	executable = os.path.abspath(sys.argv[0])
	self.trace('Executable "%s" as node process (%d)' % (executable, os.getpid()))
	self.trace('Working folder "%s"' % (os.getcwd()))
	self.trace('Running object "%s"' % (object_type.__art__.path,))
	self.trace('Class threads (%d) %s' % (len(ap.pt.thread_classes), ','.join(name_counts)))

	# One source of directory information.
	# Persistent.
	p = role.properties

	def return_signal(value):
		ar.co.signal_received = signal.SIGKILL
		return value

	# Start with the scope enumeration passed through from
	# create_node().
	scope = node_settings.node_scope
	if scope is None:
		return ar.Failed(node_scope=(None, 'scope is undefined'))

	if scope == ScopeOfService.PROCESS:
		group_port = node_settings.group_port
		if group_port is None:
			connect_above = HostPort()
			self.trace('No group port available')
		else:
			connect_above = LocalPort(group_port)
			self.trace(f'Detected group port {connect_above}')
		
		accept_below = HostPort(host=None)				# Null. Disabled.

		a = self.create(ServiceDirectory, scope, connect_above, accept_below)
		pb.directory = a

		# Wait for operational directory, esp. ephemeral.
		m = self.select(HostPort, ar.Completed, ar.Stop, ar.Faulted)
		if isinstance(m, ar.Completed):
			return return_signal(m.value)
		elif isinstance(m, ar.Stop):
			# Directory stopped in AddOn.
			return return_signal(ar.Aborted())
		elif isinstance(m, ar.Faulted):
			return return_signal(m)

	pa = ()
	if settings is not None:
		pa = pa + (settings,)
	if input is not None:
		pa = pa + (input,)
	if variables is not None:
		pa = pa + (variables,)
	pa = pa + fixed_value

	a = self.create(object_type, *pa, **key_value)

	try:
		while True:
			m = self.select(ar.Completed, ar.Stop, ar.Pause, ar.Resume)

			if isinstance(m, ar.Completed):
				# Do a "fake" signaling. Sidestep all the platform machinery
				# and just set a global. It does avoid any complexities
				# arising from overlapping events. Spent far too much time
				# trying to untangle signals, exceptions and interrupted i/o.
				ar.co.signal_received = signal.SIGKILL
				return m.value
			elif isinstance(m, ar.Stop):
				# Received a Stop.
				self.send(m, a)
				m = self.select(ar.Completed)
				return m.value
			
			self.send(m, a)
	finally:
		pass

ar.bind_function(node_vector, lifecycle=True, message_trail=True, execution_trace=True)

#
#
def create_node(object_type, *fixed_value,
	factory_settings=None, factory_input=None, factory_variables=None,
	parameter_passing=node_passing, parameter_table=None,
	upgrade=None, logs=ar.log_to_nowhere,
	scope=ScopeOfService.PROCESS, **key_value):
	"""Creates an async pub-sub process shim around a "main" async object. Returns nothing.

	:param object_type: the type of an async object to be instantiated
	:type object_type: a function or a Point-based class
	:param fixed_value: position args to forward to the object
	:type fixed_value: tuple
	:param factory_settings: persistent values
	:type factory_settings: instance of a registered class
	:param factory_input: per-invocation values
	:type factory_input: instance of a registered class
	:param factory_variables: host environment values
	:type factory_variables: instance of a registered class
	:param parameter_passing: method for parsing sys.argv[]
	:type parameter_passing: a function
	:param parameter_table: table of sub-commands and their associated functions
	:type parameter_table: dict
	:param upgrade: function that accepts old versions of settings/input and produces the current version
	:type upgrade: function
	:param logs: a callable object expecting to receive log objects
	:type logs: function or class with __call__ method
	:param scope: level of the internal directory object
	:type scope: enumeration
	:param key_value: named args to forward to the object
	:type key_value: dict
	:rtype: None
	"""
	node_settings.node_scope = scope

	ar.create_object(object_type, *fixed_value,
		factory_settings=factory_settings, factory_input=factory_input, factory_variables=factory_variables,
		parameter_passing=parameter_passing, parameter_table=parameter_table,
		start_vector=node_vector,
		upgrade=upgrade, logs=logs, properties=NodeProperties, **key_value)



# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
__docformat__ = 'restructuredtext'

import ansar.create as ar
from .socketry import *
from .transporting import *
from .plumbing import *
from .directory_if import *
from .directory import *
from .networking_if import *

__all__ = [
	'ConnectToAddress',
	'ListenAtAddress',
	'SubscribeToListing',
	'PublishAListing',
	'SubscribeToSearch',
]

#
#
class INITIAL: pass
class PENDING: pass
class CONNECTED: pass
class GLARING: pass
class CLOSING: pass
class ACCEPTING: pass
class READY: pass

#
#
class ConnectToAddress(ar.Point, ar.StateMachine):
	"""Maintain a connection to an IP address and port, perform retries.

	:param ipp: IP address and port to connect to
	:type ipp: HostPort
	:param keep_connected: restart retry sequence after successful connect
	:type keep_connected: bool
	:param session: object to be created on connection, or None
	:type session: CreateFrame
	:param group_address: where to forward session messages, or None
	:type group_address: async address
	"""
	def __init__(self, ipp, keep_connected=True, session=None, group_address=None):
		ar.Point.__init__(self)
		ar.StateMachine.__init__(self, INITIAL)
		self.ipp = ipp
		self.keep_connected = keep_connected
		self.session = session
		self.group_address = group_address

		self.started = None
		self.attempts = 0

		self.connected = None
		self.remote = None

		self.closing = None
		self.intervals = None
		self.retry = None

	def reschedule(self):
		if self.intervals is None:
			s = local_private_public(self.ipp.host)
			r = ip_retry(s)
			self.intervals = r
		
		if self.retry is None:
			self.retry = ar.smart_intervals(self.intervals)

		try:
			p = next(self.retry)
		except StopIteration:
			self.retry = None
			return False
	
		self.start(GlareTimer, p)
		return True

# INITIAL
# Launch this object.
def ConnectToAddress_INITIAL_Start(self, message):
	self.group_address = self.group_address or self.parent_address

	# Start from nothing.
	self.started = ar.world_now()
	connect(self, self.ipp, session=self.session)
	self.attempts = 1
	return PENDING

# PENDING
# Waiting for results of connect.
# Transport established.
def ConnectToAddress_PENDING_Connected(self, message):
	self.connected = message
	self.remote = self.return_address

	# Remote object is ready.
	self.send(UseAddress(self.return_address), self.parent_address)
	return CONNECTED

def ConnectToAddress_PENDING_NotConnected(self, message):
	# Attempt failed.
	# No session and no change of status for owner.
	# Schedule another or perhaps end of attempts.
	if self.reschedule():
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ConnectToAddress_PENDING_Stop(self, message):
	# Local termination.
	# Connected could be orphaned here.
	self.complete(ar.Aborted())

# CONNECTED
# Caretaker role. Pass app messages on to owner
# and wait for further control messages.
def ConnectToAddress_CONNECTED_Unknown(self, message):
	# Normal operation.	Forward app message on to proper target.
	self.forward(message, self.group_address, self.return_address)
	return CONNECTED

def ConnectToAddress_CONNECTED_Abandoned(self, message):
	# Normal end of a session.
	# Are there intended to be others?
	if self.keep_connected:
		# Start the retries up again.
		self.started = ar.world_now()
		self.attempts = 0
		self.retry = None
		if self.reschedule():
			# Update the owner that the current session
			# is over.
			self.send(NoAddress(), self.parent_address)
			return GLARING
		# Will only happen on a retry value that
		# allows no retries.
		x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
		self.complete(x)

	# End of session and only wanted 1.
	self.complete(message)

def ConnectToAddress_CONNECTED_Stop(self, message):
	# This object ended by app. Take that as
	# signal to end this session and not activate retries.

	e = ar.Stop() if self.session else Close(ar.Aborted())

	self.send(e, self.remote)
	return CLOSING

def ConnectToAddress_CONNECTED_Closed(self, message):
	# Local end has sent close to the proxy. Treat this
	# as a short-circuit version of above.
	self.complete(message.value)

# GLARING
# After a failed attempt or after abandoned.
def ConnectToAddress_GLARING_Unknown(self, message):
	# Non-control message sneaking through.
	self.forward(message, self.group_address, self.return_address)
	return GLARING

def ConnectToAddress_GLARING_GlareTimer(self, message):
	connect(self, self.ipp, session=self.session)
	self.attempts += 1
	return PENDING

def ConnectToAddress_GLARING_Stop(self, message):
	# Drop GlareTimer
	self.complete(ar.Aborted())

# CLOSING
def ConnectToAddress_CLOSING_Unknown(self, message):
	# Normal operation.	Forward app message on to proper target.
	self.forward(message, self.group_address, self.return_address)
	return CLOSING

def ConnectToAddress_CLOSING_Abandoned(self, message):
	# Terminated by remote before close could get through.
	self.complete(message)

def ConnectToAddress_CLOSING_Closed(self, message):
	# Completion of CONNECTED-Stop.
	self.complete(message.value)

CONNECT_TO_ADDRESS_DISPATCH = {
	INITIAL: (
		(ar.Start,), ()
	),
	PENDING: (
		(Connected, NotConnected, ar.Stop), ()
	),
	CONNECTED: (
		(ar.Unknown, Abandoned, ar.Stop, Closed), ()
	),
	GLARING: (
		(ar.Unknown, GlareTimer, ar.Stop), ()
	),
	CLOSING: (
		(ar.Unknown, Abandoned, Closed), ()
	),
}

ar.bind(ConnectToAddress, CONNECT_TO_ADDRESS_DISPATCH)

#
#
LISTEN_RETRY = ar.RetryIntervals(first_steps=[], regular_steps=4.0, randomized=0.25)

class ListenAtAddress(ar.Point, ar.StateMachine):
	"""Maintain a network presence at an IP address and port.

	:param ipp: IP address and port to listen at
	:type ipp: HostPort
	:param session: object to be created on accept, or None
	:type session: CreateFrame
	:param group_address: where to forward session messages, or None
	:type group_address: async address
	"""
	def __init__(self, ipp, session=None, group_address=None):
		ar.Point.__init__(self)
		ar.StateMachine.__init__(self, INITIAL)
		self.ipp = ipp
		self.session = session
		self.group_address = group_address

		self.started = None
		self.attempts = 0

		self.listening = None
		self.accepted = {}

		self.closing = None
		self.retry = None

	def reschedule(self):
		if self.retry is None:
			self.retry = ar.smart_intervals(LISTEN_RETRY)

		try:
			p = next(self.retry)
		except StopIteration:
			self.retry = None
			return False
	
		self.start(GlareTimer, p)
		return True

# INITIAL
# Launch this object.
def ListenAtAddress_INITIAL_Start(self, message):
	self.group_address = self.group_address or self.parent_address

	# Start from nothing.
	self.started = ar.world_now()
	listen(self, self.ipp, session=self.session)
	self.attempts = 1
	return PENDING

# PENDING
# Waiting for results of connect.
# Transport established.
def ListenAtAddress_PENDING_Listening(self, message):
	self.listening = message
	self.remote = self.return_address

	# Ready to accept
	self.send(UseAddress(self.address), self.parent_address)
	return ACCEPTING

def ListenAtAddress_PENDING_NotListening(self, message):
	# Attempt failed.
	# No session and no change of status for owner.
	# Schedule another or perhaps end of attempts.
	if self.reschedule():
		return GLARING

	x = ar.Exhausted(message, attempts=self.attempts, started=self.started)
	self.complete(x)

def ListenAtAddress_PENDING_Stop(self, message):
	# Local termination.
	# Listening could be orphaned here.
	self.complete(ar.Aborted())

# ACCEPTING
# Caretaker role. Pass app messages on to owner
# and wait for further control messages.
def ListenAtAddress_ACCEPTING_Accepted(self, message):
	# Start of a session.
	# Are there intended to be others?
	self.accepted[self.return_address] = message

	self.forward(message, self.parent_address, self.return_address)
	return ACCEPTING

def ListenAtAddress_ACCEPTING_Unknown(self, message):
	# Normal operation.	Forward app message on to proper target.
	self.forward(message, self.group_address, self.return_address)
	return ACCEPTING

def ListenAtAddress_ACCEPTING_Abandoned(self, message):
	a = self.accepted.pop(self.return_address, None)
	if a:
		self.forward(message, self.parent_address, self.return_address)
	return ACCEPTING

def ListenAtAddress_ACCEPTING_Closed(self, message):
	a = self.accepted.pop(self.return_address, None)
	if a:
		self.forward(message, self.parent_address, self.return_address)
	return ACCEPTING

def ListenAtAddress_ACCEPTING_Stop(self, message):
	if len(self.accepted) == 0:
		self.complete(ar.Aborted())

	e = ar.Stop() if self.session else Close(ar.Aborted())

	for k, v in self.accepted.items():
		self.send(e, k)
	return CLOSING

# GLARING
# After a failed attempt or after abandoned.
def ListenAtAddress_GLARING_Unknown(self, message):
	# Non-control message sneaking through.
	self.forward(message, self.group_address, self.return_address)
	return GLARING

def ListenAtAddress_GLARING_GlareTimer(self, message):
	listen(self, self.ipp, session=self.session)
	self.attempts += 1
	return PENDING

def ListenAtAddress_GLARING_Stop(self, message):
	# Drop GlareTimer
	self.complete(ar.Aborted())

# CLOSING
def ListenAtAddress_CLOSING_Abandoned(self, message):
	a = self.accepted.pop(self.return_address, None)
	if a and len(self.accepted) > 0:
		return CLOSING
	self.complete(message)

def ListenAtAddress_CLOSING_Closed(self, message):
	a = self.accepted.pop(self.return_address, None)
	if a and len(self.accepted) > 0:
		return CLOSING
	self.complete(message)

LISTEN_AT_ADDRESS_DISPATCH = {
	INITIAL: (
		(ar.Start,), ()
	),
	PENDING: (
		(Listening, NotListening, ar.Stop), ()
	),
	ACCEPTING: (
		(Accepted, ar.Unknown, Abandoned, Closed, ar.Stop), ()
	),
	GLARING: (
		(ar.Unknown, GlareTimer, ar.Stop), ()
	),
	CLOSING: (
		(Abandoned, Closed), ()
	),
}

ar.bind(ListenAtAddress, LISTEN_AT_ADDRESS_DISPATCH)

#
# Subscribe/Publish
class SubscribeToListing(ar.Point, ar.StateMachine):
	"""Maintain a connection to a published name.

	:param listing: the name of interest (not a regular expression)
	:type listing: str
	:param scope: highest level to look for a match
	:type scope: enumeration
	:param keep_connected: allow for multiple sessions
	:type keep_connected: bool
	:param session: object to be created on connection, or None
	:type session: CreateFrame
	:param group_address: where to forward session messages, or None
	:type group_address: async address
	"""
	def __init__(self, listing, scope=ScopeOfService.WAN, keep_connected=True, session=None, group_address=None):
		ar.Point.__init__(self)
		ar.StateMachine.__init__(self, INITIAL)
		self.listing = listing
		self.scope = scope
		self.keep_connected = keep_connected
		self.session = session
		self.group_address = group_address

		self.started = None
		self.attempts = 0

		self.subscribed = None
		self.available = None
		self.remote = None

		self.closing = None
		self.intervals = None
		self.retry = None

	def reschedule(self):
		if self.intervals is None:
			s = local_private_public(self.ipp.host)
			r = ip_retry(s)
			self.intervals = r
		
		if self.retry is None:
			self.retry = ar.smart_intervals(self.intervals)

		try:
			p = next(self.retry)
		except StopIteration:
			self.retry = None
			return False
	
		self.start(GlareTimer, p)
		return True

# INITIAL
# Launch this object.
def SubscribeToListing_INITIAL_Start(self, message):
	self.group_address = self.group_address or self.parent_address

	# Start from nothing.
	self.started = ar.world_now()
	subscribe(self, self.listing, create_session=self.session, requested_scope=self.scope)
	self.attempts = 1
	return PENDING

# PENDING
# Waiting for results of subscibe.
# Subscription acknowledged.
def SubscribeToListing_PENDING_Subscribed(self, message):
	self.subscribed = message
	return READY

def SubscribeToListing_PENDING_NotSubscribed(self, message):
	self.complete(message)

def SubscribeToListing_PENDING_Stop(self, message):
	# Local termination.
	# Subscribed could be orphaned here.
	self.complete(ar.Aborted())

# READY
# Waiting for availability.
# Subscription acknowledged.
def SubscribeToListing_READY_Available(self, message):
	self.available = message
	self.remote = self.return_address

	# Remote object is ready.
	self.send(UseAddress(self.return_address), self.parent_address)
	return CONNECTED

def SubscribeToListing_READY_Stop(self, message):
	retract(self)
	self.complete(ar.Aborted())

# CONNECTED
# Caretaker role. Pass app messages on to owner
# and wait for further control messages.
def SubscribeToListing_CONNECTED_Unknown(self, message):
	# Normal operation.	Forward app message on to proper target.
	self.forward(message, self.group_address, self.return_address)
	return CONNECTED

def SubscribeToListing_CONNECTED_Dropped(self, message):
	# Normal end of a session.
	# Are there intended to be others?
	self.send(NoAddress(), self.parent_address)
	if self.keep_connected:
		return READY

	# End of session and only wanted 1.
	self.complete(message)

def SubscribeToListing_CONNECTED_Cleared(self, message):
	# Normal end of a session.
	# Are there intended to be others?
	self.send(NoAddress(), self.parent_address)
	if self.keep_connected:
		return READY

	# End of session and only wanted 1.
	self.complete(message.value)

def SubscribeToListing_CONNECTED_Stop(self, message):
	# This object ended by app. Take that as
	# signal to end this session and not activate retries.

	self.complete(ar.Aborted())

	#e = ar.Stop() if self.session else Close(ar.Aborted())

	#self.send(e, self.remote)
	#return CLOSING

# CLOSING
def SubscribeToListing_CLOSING_Unknown(self, message):
	# Normal operation.	Forward app message on to proper target.
	self.forward(message, self.group_address, self.return_address)
	return CLOSING

def SubscribeToListing_CLOSING_Abandoned(self, message):
	# Terminated by remote before close could get through.
	retract(self)
	self.complete(message)

def SubscribeToListing_CLOSING_Closed(self, message):
	# Completion of CONNECTED-Stop.
	retract(self)
	self.complete(message.value)

SUBSCRIBE_TO_LISTING_DISPATCH = {
	INITIAL: (
		(ar.Start,), ()
	),
	PENDING: (
		(Subscribed, ar.Stop), ()
	),
	READY: (
		(Available, ar.Stop), ()
	),
	CONNECTED: (
		(ar.Unknown, Dropped, Cleared, ar.Stop), ()
	),
	CLOSING: (
		(ar.Unknown, Abandoned, Closed), ()
	),
}

ar.bind(SubscribeToListing, SUBSCRIBE_TO_LISTING_DISPATCH)

#
#
class PublishAListing(ar.Point, ar.StateMachine):
	"""Maintain a network presence at a name.

	:param listing: the name this object will be known by
	:type listing: str
	:param session: object to be created on delivery, or None
	:type session: CreateFrame
	:param group_address: where to forward session messages, or None
	:type group_address: async address
	"""
	def __init__(self, listing, scope=ScopeOfService.WAN, session=None, group_address=None):
		ar.Point.__init__(self)
		ar.StateMachine.__init__(self, INITIAL)
		self.listing = listing
		self.scope = scope
		self.session = session
		self.group_address = group_address

		self.started = None
		self.attempts = 0

		self.published = None
		self.delivered = {}

		self.closing = None
		self.retry = None

	def reschedule(self):
		if self.retry is None:
			self.retry = ar.smart_intervals(LISTEN_RETRY)

		try:
			p = next(self.retry)
		except StopIteration:
			self.retry = None
			return False
	
		self.start(GlareTimer, p)
		return True

# INITIAL
# Launch this object.
def PublishAListing_INITIAL_Start(self, message):
	self.group_address = self.group_address or self.parent_address

	# Start from nothing.
	self.started = ar.world_now()
	publish(self, self.listing, create_session=self.session, requested_scope=self.scope)
	self.attempts = 1
	return PENDING

# PENDING
# Waiting for results of publish.
# Listing established.
def PublishAListing_PENDING_Published(self, message):
	self.published = message

	# Ready to accept
	self.send(UseAddress(self.address), self.parent_address)
	return ACCEPTING

def PublishAListing_PENDING_NotPublished(self, message):
	self.complete(message)

def PublishAListing_PENDING_Stop(self, message):
	# Local termination.
	retract(self)
	self.complete(ar.Aborted())

# ACCEPTING
# Caretaker role. Pass app messages on to owner
# and wait for further control messages.
def PublishAListing_ACCEPTING_Delivered(self, message):
	# Start of a session.
	# Are there intended to be others?
	self.delivered[self.return_address] = message

	self.forward(message, self.parent_address, self.return_address)
	return ACCEPTING

def PublishAListing_ACCEPTING_Unknown(self, message):
	# Normal operation.	Forward app message on to proper target.
	self.forward(message, self.group_address, self.return_address)
	return ACCEPTING

def PublishAListing_ACCEPTING_Dropped(self, message):
	a = self.delivered.pop(self.return_address, None)
	if a:
		self.forward(message, self.parent_address, self.return_address)
	return ACCEPTING

def PublishAListing_ACCEPTING_Cleared(self, message):
	a = self.delivered.pop(self.return_address, None)
	if a:
		self.forward(message, self.parent_address, self.return_address)
	return ACCEPTING

def PublishAListing_ACCEPTING_Stop(self, message):
	retract(self)
	self.complete(ar.Aborted())
	#if len(self.delivered) == 0:
	#	retract(self)
	#	self.complete(ar.Aborted())
	#
	#e = ar.Stop() if self.session else Close(ar.Aborted())
	#
	#for k, v in self.delivered.items():
	#	self.send(e, k)
	#return CLOSING

# CLOSING
def PublishAListing_CLOSING_Dropped(self, message):
	a = self.delivered.pop(self.return_address, None)
	if a and len(self.delivered) > 0:
		return CLOSING
	retract(self)
	self.complete(message)

def PublishAListing_CLOSING_Cleared(self, message):
	a = self.delivered.pop(self.return_address, None)
	if a and len(self.delivered) > 0:
		return CLOSING
	retract(self)
	self.complete(message)

PUBLISH_A_LISTING_DISPATCH = {
	INITIAL: (
		(ar.Start,), ()
	),
	PENDING: (
		(Published, NotPublished, ar.Stop), ()
	),
	ACCEPTING: (
		(Delivered, ar.Unknown, Dropped, Cleared, ar.Stop), ()
	),
	CLOSING: (
		(Dropped, Cleared), ()
	),
}

ar.bind(PublishAListing, PUBLISH_A_LISTING_DISPATCH)

#
#
class SubscribeToSearch(ar.Point, ar.StateMachine):
	"""Maintain connections with multiple published names.

	:param listing: the names to search for (regular expression)
	:type listing: str
	:param scope: highest level to look for a match
	:type scope: enumeration
	:param session: object to be created on connection, or None
	:type session: CreateFrame
	:param group_address: where to forward session messages, or None
	:type group_address: async address
	"""
	def __init__(self, search, scope=ScopeOfService.WAN, session=None, group_address=None):
		ar.Point.__init__(self)
		ar.StateMachine.__init__(self, INITIAL)
		self.search = search
		self.scope = scope
		self.session = session
		self.group_address = group_address

		self.started = None
		self.attempts = 0

		self.subscribed = None
		self.available = {}

		self.closing = None
		self.retry = None

	def reschedule(self):
		if self.retry is None:
			self.retry = ar.smart_intervals(LISTEN_RETRY)

		try:
			p = next(self.retry)
		except StopIteration:
			self.retry = None
			return False
	
		self.start(GlareTimer, p)
		return True

# INITIAL
# Launch this object.
def SubscribeToSearch_INITIAL_Start(self, message):
	self.group_address = self.group_address or self.parent_address

	# Start from nothing.
	self.started = ar.world_now()
	subscribe(self, self.search, create_session=self.session, requested_scope=self.scope)
	self.attempts = 1
	return PENDING

# PENDING
# Waiting for results of publish.
# Listing established.
def SubscribeToSearch_PENDING_Subscribed(self, message):
	self.subscribed = message

	# Ready to accept
	self.send(UseAddress(self.address), self.parent_address)
	return ACCEPTING

def SubscribeToSearch_PENDING_NotSubscribed(self, message):
	self.complete(message)

def SubscribeToSearch_PENDING_Stop(self, message):
	# Local termination.
	retract(self)
	self.complete(ar.Aborted())

# ACCEPTING
# Caretaker role. Pass app messages on to owner
# and wait for further control messages.
def SubscribeToSearch_ACCEPTING_Available(self, message):
	# Start of a session.
	# Are there intended to be others?
	self.available[self.return_address] = message

	self.forward(message, self.parent_address, self.return_address)
	return ACCEPTING

def SubscribeToSearch_ACCEPTING_Unknown(self, message):
	# Normal operation.	Forward app message on to proper target.
	self.forward(message, self.group_address, self.return_address)
	return ACCEPTING

def SubscribeToSearch_ACCEPTING_Dropped(self, message):
	a = self.available.pop(self.return_address, None)
	if a:
		self.forward(message, self.parent_address, self.return_address)
	return ACCEPTING

def SubscribeToSearch_ACCEPTING_Cleared(self, message):
	a = self.available.pop(self.return_address, None)
	if a:
		self.forward(message, self.parent_address, self.return_address)
	return ACCEPTING

def SubscribeToSearch_ACCEPTING_Stop(self, message):
	retract(self)
	self.complete(ar.Aborted())
	#if len(self.available) == 0:
	#	retract(self)
	#	self.complete(ar.Aborted())
	#
	#e = ar.Stop() if self.session else Close(ar.Aborted())
	#
	#for k, v in self.available.items():
	#	self.send(e, k)
	#return CLOSING

# CLOSING
def SubscribeToSearch_CLOSING_Dropped(self, message):
	a = self.available.pop(self.return_address, None)
	if a and len(self.available) > 0:
		return CLOSING
	retract(self)
	self.complete(message)

def SubscribeToSearch_CLOSING_Cleared(self, message):
	a = self.available.pop(self.return_address, None)
	if a and len(self.available) > 0:
		return CLOSING
	retract(self)
	self.complete(message)

SUBSCRIBE_TO_SEARCH_DISPATCH = {
	INITIAL: (
		(ar.Start,), ()
	),
	PENDING: (
		(Subscribed, ar.Stop), ()	# NotSubscribed?!
	),
	ACCEPTING: (
		(Available, ar.Unknown, Dropped, Cleared, ar.Stop), ()
	),
	CLOSING: (
		(Dropped, Cleared), ()
	),
}

ar.bind(SubscribeToSearch, SUBSCRIBE_TO_SEARCH_DISPATCH)



# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2024 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
'''.

.
'''
__docformat__ = 'restructuredtext'

import ansar.create as ar
from .socketry import *
from .transporting import *
from .plumbing import *
from .networking_if import *

__all__ = [
	'GroupTable',
	'GroupUpdate',
	'AddressGroup',
	'GroupObject',
]

# The specification.
class GroupTable(object):
	"""Table of objects that create runtime objects, e.g. manage multiple connections.

	:param member_frame: list of named CreateFrames
	:type member_frame: dict
	"""
	def __init__(self, **member_frame):
		if 'member_frame' in member_frame or 'create' in member_frame or 'update' in member_frame:
			raise ValueError('names of members would hide GroupTable methods')
		self.member_frame = member_frame

	def create(self, owner, get_ready=None, session=None, resident_code=False, until_stopped=False):
		"""Create a child instance of `AddressGroup`. Return the address of the new object.

		:param owner: async object to receive GroupUpdate messages
		:type owner: Point
		:param get_ready: time limit on achieving the ready state, or None
		:type get_ready: float
		:param session: object to be created at Ready state, or None
		:type session: CreateFrame
		:param resident_code: facilitate a series of sessions
		:type resident_code: bool
		:param until_stopped: a flag passed to AddressGroup
		:type until_stopped: bool
		:rtype: ansar address
		"""
		for k in self.member_frame.keys():
			setattr(self, k, None)	# Fill with blanks.
		a = owner.create(AddressGroup, self.member_frame, session=session, resident_code=resident_code, get_ready=get_ready, until_stopped=until_stopped)
		return a

	def update(self, message):
		"""Process a GroupUpdate message from the AddressGroup object.

		:param message: acquire a new address or lose an existing address
		:type message: GroupUpdate
		:rtype: None
		"""
		if message.key in self.member_frame:
			setattr(self, message.key, message.address)

# Runtime image.
class GroupUpdate(object):
	"""A group notification, an address was acquired or lost.

	:param key: name of the address
	:type key: str
	:param address: new address or None
	:type address: ansar address
	"""
	def __init__(self, key=None, address=None):
		self.key = key
		self.address = address

UPDATE_GROUP_SCHEMA = {
	'key': ar.Unicode(),
	'address': ar.Address(),
}

ar.bind(GroupUpdate, object_schema=UPDATE_GROUP_SCHEMA)

# Dedicated timers.
class GroupTimer(object): pass

ar.bind(GroupTimer)

class Group(object): pass

#
#
class INITIAL: pass
class PENDING: pass
class LATCHING: pass
class READY: pass
class RESTARTING: pass
class UNLATCHING: pass
class CLEARING: pass
class RUNNING: pass

LATCH_ID = 1
SESSION_ID = 2

class AddressGroup(ar.Point, ar.StateMachine):
	"""Manage a collection of objects that are acquiring addresses, e.g. network connections.

	:param table: table of named CreateFrames
	:type table: dict
	:param session: object to be created at ready state, or None
	:type session: CreateFrame
	:param resident_code: facilitate a series of sessions
	:type resident_code: bool
	:param get_ready: time limit on achieving the ready state, or None
	:type get_ready: float
	:param until_stopped: override the completion value with the session completion
	:type until_stopped: bool
	"""
	def __init__(self, table, session=None, resident_code=False, get_ready=None, until_stopped=False):
		ar.Point.__init__(self)
		ar.StateMachine.__init__(self, INITIAL)
		self.table = table						# Rows of CreateFrames.
		self.session = session					# The object to be created or null.
		self.resident_code = resident_code		# Is this an endless series of sessions.
		self.get_ready = get_ready				# Limit the setup time.
		self.until_stopped = until_stopped		# Termination by intervention, i.e. command?

		self.latch = None		# Temporary address while pending.
		self.created = None		# Address of the session.

		self.ready = Group()			# The set of runtime addresses.
		self.return_value = None		# Hold the value for final complete.
		self.session_value = None		# Pass completion value from one session to the next.


# START
# Start each of the member objects. If appropriate, create a
# temporary address to capture messages from objects that are
# ready before the group is ready.
def AddressGroup_INITIAL_Start(self, message):
	if not self.table:
		self.complete(ar.Faulted('empty table', 'a group with no members has nothing to do'))

	if not self.resident_code and self.get_ready is not None:
		self.start(GroupTimer, self.get_ready)

	if self.session:
		self.latch = self.create(ar.Latch, '', self.parent_address)
		self.assign(self.latch, LATCH_ID)

	for k, v in self.table.items():
		setattr(self.ready, k, None)
		a = self.create(v.object_type, *v.args, group_address=self.latch, **v.kw)
		self.assign(a, k)
	
	return PENDING

# PENDING
# Assumption is that eventually all members will provide a runtime
# address and the whole group will go ready. Which is the moment
# a session is created.
def AddressGroup_PENDING_UseAddress(self, message):
	k = self.progress()
	if k is None:
		self.warning(f'Unknown sender {self.return_address}')
		return PENDING

	a = getattr(self.ready, k, None)
	if a:
		self.warning(f'Table entry "{k}" already on record (ignored)')
		return PENDING

	setattr(self.ready, k, message.address)
	def ready():
		for k in self.table.keys():
			a = getattr(self.ready, k, None)
			if a is None:
				return False
		return True

	r = ready()
	if not self.session:
		# Owner expects notification stream. Add the
		# new address and provide trailing edge of
		# READY event.
		self.send(GroupUpdate(k, message.address), self.parent_address)
		if r: 
			self.send(ar.Ready(), self.parent_address)
			return READY
		return PENDING

	if r:
		# Use the factory to create a new session object passing the
		# object of addresses. Use distinct call signature for create()
		# to avoid imposing session_value=None on every group object.
		c = self.session
		if self.session_value is None:
			a = self.create(c.object_type, self.ready, *c.args, **c.kw)
		else:
			a = self.create(c.object_type, self.ready, *c.args, session_value=self.session_value, **c.kw)
		self.assign(a, SESSION_ID)
		self.created = a
		# Get latch to work the address magic so that session
		# receives potentially crucial messages from members.
		self.send(ar.SwitchOver(a), self.latch)
		return LATCHING

	return PENDING

def AddressGroup_PENDING_NoAddress(self, message):
	k = self.progress()
	if k is None:
		self.warning(f'Unknown sender {self.return_address}')
		return PENDING

	a = getattr(self.ready, k, None)
	if a is None:
		self.warning(f'Table entry "{k}" was not on record')
		return PENDING

	setattr(self.ready, k, None)

	if not self.session:
		self.send(GroupUpdate(k, None), self.parent_address)

	return PENDING

def AddressGroup_PENDING_Completed(self, message):
	k = self.debrief()
	if k is None:
		self.warning(f'Unknown sender {self.return_address}')
		return PENDING

	# All permutations of group terminate if any member
	# drops out.
	if isinstance(k, str):
		a = getattr(self.ready, k, None)

		if not self.session and a:
			self.send(GroupUpdate(k, None), self.parent_address)

	if self.working():
		self.return_value = message.value
		self.abort()
		return CLEARING

	self.complete(message.value)

def AddressGroup_PENDING_GroupTimer(self, message):
	self.return_value = ar.TimedOut(message)
	self.abort()
	return CLEARING

def AddressGroup_PENDING_Stop(self, message):
	self.return_value = ar.Aborted()
	self.abort()
	return CLEARING

# LATCHING
# The group is going ready, just need confirmation
# that latching has worked. Use that message to
# rewire the "assigned job" machinery.
def AddressGroup_LATCHING_Ack(self, message):
	j = self.address_job.pop(self.latch, None)
	if j != LATCH_ID:
		self.return_value = ar.Faulted('Ack but not the assigned latch', 'internal')
		self.warning(f'{self.group_value}')
		self.abort()
		return CLEARING
	self.assign(self.return_address, LATCH_ID)		# Latch has been re-homed to this new address.
	self.latch = self.return_address
	return READY

def AddressGroup_LATCHING_Unknown(self, message):
	self.save(message)
	return LATCHING

def AddressGroup_LATCHING_Stop(self, message):
	self.return_value = ar.Aborted()
	self.abort()
	return CLEARING

# READY
# The session object is running, presumably sending and
# receiving from the group of runtime messages.
def AddressGroup_READY_NoAddress(self, message):
	k = self.progress()
	if k is None:
		self.warning(f'Unknown sender {self.return_address}')
		return READY

	a = getattr(self.ready, k, None)
	if a is None:
		self.warning(f'Table entry "{k}" was not on record')
		return READY

	# Lost the READY status.
	setattr(self.ready, k, None)

	if not self.session:
		# Owner wants notication stream.
		# Delete the named address and provide leading
		# edge "not ready" event. Back to PENDING state.
		self.send(ar.NotReady(), self.parent_address)
		self.send(GroupUpdate(k, None), self.parent_address)
		return PENDING

	if self.resident_code:
		# This group is here for the long run. Terminate
		# the session.
		self.send(ar.Stop(), self.created)
		return RESTARTING

	# This code terminates at the end of the ready state.
	if self.working():
		self.return_value = ar.Faulted(f'Lost address "{k}" in group session', 'group no longer complete')
		self.abort()
		return CLEARING

	self.complete(message)

def AddressGroup_READY_Completed(self, message):
	k = self.debrief()
	if k is None:
		self.warning(f'Unknown address {self.return_address}')
		return PENDING

	if isinstance(k, str):
		if not self.session:
			# Leading edge notification.
			self.send(ar.NotReady(), self.parent_address)
			self.send(GroupUpdate(k, None), self.parent_address)
		self.return_value = message.value
	elif k == SESSION_ID:
		self.return_value = message.value
	elif k == LATCH_ID:
		self.return_value = message.value
	else:
		self.return_value = ar.Faulted(f'Address "{k}" completed and not member or id', 'internal')

	if self.working():
		self.abort()
		return CLEARING

	self.complete(self.return_value)

def AddressGroup_READY_Unknown(self, message):
	# Probably application message that needs to be
	# forwarded onto the proper object, i.e. the active
	# session of the owner of the group.
	if self.created:
		self.forward(message, self.created, self.return_address)
	else:
		self.forward(message, self.parent_address, self.return_address)
	return READY

def AddressGroup_READY_Stop(self, message):
	self.return_value = ar.Aborted()
	self.abort()
	return CLEARING

# RESTARTING
# Stopped the session and waiting for its
# termination.
def AddressGroup_RESTARTING_Completed(self, message):
	j = self.debrief()
	if j != SESSION_ID:
		self.return_value = ar.Faulted(f'Lost member "{j}" during restart', 'shift to termination')
		if self.working():
			self.abort()
			return CLEARING
		self.complete(self.group_value)

	# Save the value and pass to any subsequent
	# sessions.
	self.session_value = message.value

	# Session gone. Ask the latch to reclaim its
	# proper place.
	self.send(ar.Reclaim(), self.latch)
	return UNLATCHING

def AddressGroup_RESTARTING_Stop(self, message):
	self.return_value = ar.Aborted()
	self.abort()
	return CLEARING

# UNLATCHING
# Lost and address, dropped the session and now
# waiting for confirmation the latch is back in
# its proper place.
def AddressGroup_UNLATCHING_Ack(self, message):
	j = self.address_job.pop(self.latch, None)
	if j != LATCH_ID:
		self.return_value = ar.Faulted('Ack but not from the assigned latch', 'internal')
		self.warning(f'{self.group_value}')
		self.abort()
		return CLEARING
	self.assign(self.return_address, LATCH_ID)		# Latch has been restored to original address.
	self.latch = self.return_address
	return PENDING

def AddressGroup_UNLATCHING_Stop(self, message):
	self.return_value = ar.Aborted()
	self.abort()
	return CLEARING

# CLEARING
# This group is going down.
def AddressGroup_CLEARING_Completed(self, message):
	k = self.debrief()
	if k is None:
		self.warning(f'Unknown address {self.return_address}')
		return CLEARING

	if isinstance(k, str):
		if not self.session:
			self.send(GroupUpdate(k, None), self.parent_address)
	elif k == SESSION_ID:
		if isinstance(self.return_value, ar.Aborted) and self.until_stopped:
			self.return_value = message.value
	if self.working():
		return CLEARING

	self.complete(self.return_value)


ADDRESS_GROUP_DISPATCH = {
	INITIAL: (
		(ar.Start,), ()
	),
	PENDING: (
		(UseAddress, NoAddress, ar.Completed, GroupTimer, ar.Stop), ()
	),
	LATCHING: (
		(ar.Ack, ar.Unknown, ar.Stop), (NoAddress, ar.Completed)
	),
	READY: (
		(NoAddress, ar.Completed, ar.Unknown, ar.Stop), ()
	),
	RESTARTING: (
		(ar.Completed, ar.Stop), (NoAddress, UseAddress)
	),
	UNLATCHING: (
		(ar.Ack, ar.Stop), (ar.Completed, NoAddress, UseAddress)
	),
	CLEARING: (
		(ar.Completed,), ()
	),
}

ar.bind(AddressGroup, ADDRESS_GROUP_DISPATCH)

# Reinstated. Too many existing uses.
# To be considered deleted.

# A shim object between create_object() and the application object.
# This is the most concise way to run an object with the support
# of one or more runtime objects, e.g. a connection to a service.
class GroupObject(ar.Point, ar.StateMachine):
	def __init__(self, *args_plus_2, get_ready=None, resident_code=False, until_stopped=False):
		ar.Point.__init__(self)
		ar.StateMachine.__init__(self, INITIAL)
		# Need to trampoline a variable list of settings, input and variables
		# onto the receiving function, i.e. group_args.
		self.args = args_plus_2[:-2]
		self.session = args_plus_2[-2]
		self.group_args = args_plus_2[-1]
		self.get_ready = get_ready
		self.resident_code = resident_code
		self.until_stopped = until_stopped

		self.group = None
		self.created = None
		self.closing = None

def GroupObject_INITIAL_Start(self, message):
	# Convert settings, input and variables into args and
	# kw suitable for the session object. This is all about
	# converting the "command-line" info its most useful
	# form.
	self.group, args, kv = self.group_args(*self.args)

	# Create a session factory.
	session = CreateFrame(self.session, *args, **kv)

	# Real work done by AddressGroup.
	# Create the group passing the factory for those moments
	# when the group goes ready.
	self.created = self.group.create(self, session=session, resident_code=self.resident_code, get_ready=self.get_ready, until_stopped=self.until_stopped)
	return RUNNING

def GroupObject_RUNNING_Completed(self, message):
	self.complete(message.value)

def GroupObject_RUNNING_Stop(self, message):
	self.send(message, self.created)
	return RUNNING

GROUP_SESSION_DISPATCH = {
	INITIAL: (
		(ar.Start,), ()
	),
	RUNNING: (
		(ar.Completed, ar.Stop), ()
	),
}

ar.bind(GroupObject, GROUP_SESSION_DISPATCH)



# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
__docformat__ = 'restructuredtext'

import ansar.create as ar
from .socketry import *
from .transporting_if import ts

__all__ = [
	'connect',
	'listen',
	'stop_listen',
]

def create_sockets(root):
	ts.sockets = root.create(SocketSelect)
	ts.channel = root.select(SocketChannel)

def stop_sockets(root):
	ts.channel.send(ar.Stop(), root.address)
	root.select(ar.Completed)

ar.AddOn(create_sockets, stop_sockets)

#
#
def connect(self, requested_ipp, session=None, encrypted=False):
	"""
	Initiates a network connection to the specified IP
	address and port number.

	:param self: async entity
	:type self: Point
	:param requested_ipp: host and port to connect to
	:type requested_ipp: HostPort
	:param session: object to create on successful connection
	:type session: CreateFrame
	:param encrypted: is the server encrypting
	:type encrypted: bool
	"""
	ts.channel.send(ConnectStream(requested_ipp=requested_ipp, create_session=session, encrypted=encrypted), self.address)

#
#
def listen(self, requested_ipp, session=None, encrypted=False):
	"""
	Establishes a network presence at the specified IP
	address and port number.

	:param self: async entity
	:type self: Point
	:param requested_ipp: host and port to listen at
	:type requested_ipp: HostPort
	:param session: object to create on successful connection
	:type session: CreateFrame
	:param encrypted: is the client encrypting
	:type encrypted: bool
	"""
	ts.channel.send(ListenForStream(requested_ipp=requested_ipp, create_session=session, encrypted=encrypted), self.address)

#
#
def stop_listen(self, requested_ipp):
	ts.channel.send(StopListening(requested_ipp), self.address)



# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2022 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Utility to maintain the operational folders and files of standard components.

Starts with the process-as-an-async-object abstraction and takes it to
the next level. There is a new abstraction that manages a collection
of one or more processes. A complete runtime environment is provided for
each process, i.e. a place for temporary files and persisted files, and
a per-process configuration. Once configured to satisfy requirements a
process can be started and stopped repeatedly using single commands.
"""
__docformat__ = 'restructuredtext'

__all__ = [
	'main',
]

import sys

import ansar.connect as ar
from ansar.connect.procedure import *
from ansar.connect.wan import FOH_HOST

# Shims between command-line and library procedures.
# Provide access to ansar-encode.
versioning_settings = ar.VersioningSettings()
compare_settings = ar.CompareSettings()
release_settings = ar.ReleaseSettings()

# Provide access to ansar-create.
create_settings = ar.CreateSettings()
add_settings = ar.AddSettings()
update_settings = ar.UpdateSettings()
delete_settings = ar.DeleteSettings()
destroy_settings = ar.DestroySettings()
list_settings = ar.ListSettings()
start_settings = ar.StartSettings()
run_settings = ar.RunSettings()
pause_settings = ar.PauseSettings()
resume_settings = ar.ResumeSettings()
stop_settings = ar.StopSettings()
log_settings = ar.LogSettings()
input_settings = ar.InputSettings()
settings_settings = ar.SettingsSettings()
set_settings = ar.SetSettings()
edit_settings = ar.EditSettings()
deploy_settings = ar.DeploySettings()
returned_settings = ar.ReturnedSettings()

network_settings = NetworkSettings()
ping_settings = PingSettings()
account_settings = AccountSettings()
directory_settings = DirectorySettings()

def versioning(self, settings, ls, word):
	'''Print a human representation of the current version info. Returns nothing.'''
	version_module = ar.word_argument(0, word)
	return ar.procedure_versioned(versioning_settings, version_module)

def compare(self, settings, ls, word):
	'''Compare the current version info against saved image of most recently released.'''
	version_module = ar.word_argument(0, word)
	release_file = ar.word_argument(1, word)
	return ar.procedure_compare(compare_settings, version_module, release_file)

def release(self, settings, ls, word):
	'''Overwrite the saved image of version info with the current info.'''
	version_module = ar.word_argument(0, word)
	release_file = ar.word_argument(1, word)
	return ar.procedure_release(release_settings, version_module, release_file)


def ls_args(ls):
	args = ['--%s=%s' % (k, v) for k, v in ls[0].items()]
	args.extend(['-%s=%s' % (k, v) for k, v in ls[1].items()])
	return args

def create(self, _, ls, word):
	path = ar.word_argument(0, word)
	return ar.procedure_create(self, create_settings, path)

def add(self, _, ls, word):
	executable = ar.word_argument(0, word)
	role = ar.word_argument(1, word)
	home = ar.word_argument(2, word)
	return ar.procedure_add(self, add_settings, executable, role, home, ls)

def update(self, settings, ls, word):
	# A list of search expressions.
	role = word

	return ar.procedure_update(self, update_settings, settings.force, role, ls_args(ls))

def delete(self, settings, ls, word):
	# Settings reuse.
	role = word
	return ar.procedure_delete(self, delete_settings, role)

def list_(self, _, ls, word):
	role = word
	return ar.list_home(self, list_settings, role)

def destroy(self, settings, ls, word):
	# Settings reuse.
	home = ar.word_argument(0, word)
	return ar.procedure_destroy(self, destroy_settings, settings.force, home)

def run(self, _, ls, word):
	role = word
	return ar.procedure_run(self, run_settings, role, ls_args(ls))

def start(self, _, ls, word):
	role = word
	return ar.procedure_start(self, start_settings, role, ls_args(ls))

def pause(self, settings, ls, word):
	role = word
	return ar.procedure_pause(self, pause_settings, settings.force, role)

def resume(self, settings, ls, word):
	group = word
	return ar.procedure_resume(self, resume_settings, settings.force, group)

def stop(self, settings, ls, word):
	group = word
	return ar.procedure_stop(self, stop_settings, group)

def status(self, _, ls, word):
	# Settings reuse.
	role = word
	return ar.procedure_status(self, list_settings, role)

def history(self, _, ls, word):
	# Settings reuse.
	role = ar.word_argument(0, word)
	home = ar.word_argument(1, word)
	return ar.procedure_history(self, list_settings, role, home)

def returned(self, _, ls, word):
	role = ar.word_argument(0, word)
	home = ar.word_argument(1, word)

	return ar.procedure_returned(self, returned_settings, role, home)

def log(self, _, ls, word):
	role = ar.word_argument(0, word)
	home = ar.word_argument(1, word)

	# Initial sanity checks and a default <begin>.
	f = [log_settings.from_, log_settings.last, log_settings.start, log_settings.back]
	c = len(f) - f.count(None)
	if c == 0:
		log_settings.back = ar.text_to_span('5m')   # Default query of last 5 mins.
	elif c != 1:
		# one of <from>, <last>, <start> or <back> is required
		r = ar.Rejected(start_mark=(None, ['from', 'last', 'start', 'back']))
		raise ar.Incomplete(r)

	t = [log_settings.to, log_settings.span, log_settings.count]
	c = len(t) - t.count(None)
	if c == 0:
		pass		# Default is query to end-of-log or end of start-stop.
	elif c != 1:
		# one of <to>, <span> or <count> is required
		r = ar.Rejected(end_mark=(None, ['to', 'span', 'count']))
		raise ar.Incomplete(r)

	return ar.procedure_log(self, log_settings, role, home)

def folder(self, _, ls, word):
	# Settings reuse.
	selected = ar.word_argument(0, word)
	role = ar.word_argument(1, word)
	home = ar.word_argument(2, word)
	return ar.procedure_folder(self, list_settings, selected, role, home)

def input_(self, settings, ls, word):
	role = ar.word_argument(0, word)
	home = ar.word_argument(1, word)
	return ar.procedure_input(self, input_settings, settings.force, role, home)

def settings(self, settings, ls, word):
	# Settings reuse.
	role = ar.word_argument(0, word)
	home = ar.word_argument(1, word)
	return ar.procedure_settings(self, settings_settings, settings.force, role, home)

def get(self, _, ls, word):
	# Positional and settings options for 3 args.
	selected = ar.word_argument(0, word)
	role = ar.word_argument(1, word)
	home = ar.word_argument(2, word)
	return ar.procedure_get(self, ar.NodeProperties, set_settings, selected, role, home)

def set_(self, settings, ls, word):
	# Convert word[] into property and role[].
	if len(word) < 2:
		e = ar.Rejected(set_property=(None, 'a <property> and at least one <role> are required'))
		raise ar.Incomplete(e)

	selected = word[0]
	role = word[1:]

	if set_settings.not_set:
		js = None
	elif set_settings.property_file:
		with open(set_settings.property_file, 'r') as f:
			js = f.read()
	else:
		js = sys.stdin.read()

	return ar.procedure_set(self, ar.NodeProperties, set_settings, settings.force, selected, role, js)

def edit(self, settings, ls, word):
	# Positional and settings options for 3 args.
	selected = ar.word_argument(0, word)
	role = ar.word_argument(1, word)
	home = ar.word_argument(2, word)

	return ar.procedure_edit(self, ar.NodeProperties, edit_settings, settings.force, selected, role, home)

def deploy(self, settings, ls, word):
	build = ar.word_argument(0, word)
	snapshot = ar.word_argument(1, word)
	home = ar.word_argument(2, word)
	return ar.procedure_deploy(self, deploy_settings, settings.force, build, snapshot, home)

def snapshot(self, settings, ls, word):
	snapshot = ar.word_argument(0, word)
	home = ar.word_argument(1, word)
	return ar.procedure_snapshot(self, deploy_settings, settings.force, snapshot, home)

#
#
def network(self, settings, ls, word):
	group = ar.word_argument(0, word)
	home = ar.word_argument(1, word)
	return procedure_network(self, network_settings, group, home)

def ping(self, settings, ls, word):
	service = ar.word_argument(0, word)
	group = ar.word_argument(1, word)
	home = ar.word_argument(2, word)
	return procedure_ping(self, ping_settings, service, group, home)

#
#
def signup(self, settings, ls, word):
	return procedure_signup(self, account_settings)

def login(self, settings, ls, word):
	return procedure_login(self, account_settings)

def account(self, settings, ls, word):
	return procedure_account(self, account_settings)

def directory(self, settings, ls, word):
	return procedure_directory(self, directory_settings)

# Bring all the functions together as a table that
# uses the function name, i.e. f.__name__ as a key.
table = ar.jump_table(
	# Access to ansar-encode.
	(versioning, versioning_settings),
	(compare, compare_settings),
	(release, release_settings),

	# CRUD of process definitions.
	(create, create_settings),
	(add, add_settings),
	(update, update_settings),
	(delete, delete_settings),
	(list_, list_settings),
	(destroy, destroy_settings),

	# CRUD of processes, i.e. instances of definitions.
	(run, run_settings),
	(start, start_settings),
	(pause, pause_settings),
	(resume, resume_settings),
	(stop, stop_settings),
	(status, list_settings),

	# Monitoring and control of existing definitions
	# and any associated instances.
	(history, list_settings),
	(returned, returned_settings),
	(log, log_settings),
	(folder, list_settings),
	(input_, input_settings),
	(settings, settings_settings),
	(get, set_settings),
	(set_, set_settings),
	(edit, edit_settings),

	# Transfer of binaries and materials to-and-from
	# repo and live home.
	(deploy, deploy_settings),
	(snapshot, deploy_settings),

	# Connect-level procedures.
	(network, network_settings),
	(ping, ping_settings),
	(signup, None),
	(login, account_settings),
	(account, account_settings),
	(directory, directory_settings),
)


#
#
class Settings(object):
	def __init__(self, force=False, cloud_ip=None, encrypted=False, login_token=None):
		'''
		* force
		Override cautionary behaviour, mostly relating to the unintended termination
		or running processes.
		* cloud_ip
		Domain name or dotted IP address.
		'''
		self.force = force
		self.cloud_ip = cloud_ip
		self.encrypted = encrypted
		self.login_token = login_token

SETTINGS_SCHEMA = {
	'force': ar.Boolean(),
	'cloud_ip': ar.Unicode(),
	'encrypted': ar.Boolean(),
	'login_token': ar.UUID(),
}

ar.bind(Settings, object_schema=SETTINGS_SCHEMA)

#
#
def ansar(self, settings):
	'''Process configuration and orchestration.

	ansar create --redirect-bin=dist
	ansar add zombie
	ansar start zombie-0
	ansar status zombie-0
	ansar log zombie-0
	ansar stop
	* create
	Print the current version information on the console.
	* add
	Compare the current version information against a saved image and report
	any issues. Any issue that might subvert the processing of versioned
	materials "in the wild" will result in a non-zero exit code.
	* update
	Update the saved image with the current version information. Subsequent
	use of the compare sub-command will be relative to the new image.

	* versioning
	Print the current version information on the console.
	* compare
	Compare the current version information against a saved image and report
	any issues. Any issue that might subvert the processing of versioned
	materials "in the wild" will result in a non-zero exit code.
	* release
	Update the saved image with the current version information. Subsequent
	use of the compare sub-command will be relative to the new image.
	* create
	Create a folder to hold the configurations of one or more processes.
	* add
	Create the configuration of a process.
	* update
	Modify the configuration settings of a process.
	* delete
	Remove the configuration of a process.
	* list
	Print a list of the currently defined processes.
	* destroy
	Remove the folder and all the process configurations it may contain.
	* run
	Execute one or more processes and wait for the result.
	* start
	Initiate one or more processes in the background.
	* pause
	Terminate one or more running processes with the intention of a restart.
	* resume
	Restart all previously paused processes.
	* stop
	Terminate one or more processes.
	* status
	Print a list of the currently running processes.
	* history
	Print the start/pause/resume/stop details of a particular process.
	* returned
	Print the value returned by the most recent completion of a process.
	* log
	Print the logs created by a specified process.
	* folder
	Print an actual runtime location relating to a specified process.
	* input
	View and modify the configured input of a process.
	* settings
	View and modify the configured settings of a process.
	* get
	View the management properties of a process.
	* set
	Modify the management properties of a process.
	* edit
	View and modify the management properties of a process.
	* deploy
	Orchestrate the update of process configurations.
	* snapshot
	Take a copy of all process runtimes.
	'''
	sub_command, ls, words = ar.object_words()
	if sub_command is None:
		return None

	# Everything lined up for execution of
	# the selected sub-command.
	try:
		output = sub_command(self, settings, ls, words)
	except ar.Incomplete as e:
		return e.value
	return output

ar.bind(ansar)

#
#
factory_settings = Settings(cloud_ip=FOH_HOST, encrypted=True)

def main():
	ar.create_object(ansar, factory_settings=factory_settings, parameter_passing=ar.sub_object_passing, parameter_table=table)

# The standard entry point. Needed for IDEs
# and debugger sessions.
if __name__ == '__main__':
	main()



# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Definition of messages for client and cloud interactions.

The protocol messages that establish wide-area transports between
pub-sub pairs. Needed here for the ansar CLI and the runtime
handshaking.
"""
__docformat__ = 'restructuredtext'

import ansar.create as ar
from .socketry import *
from .product import *

__all__ = [
	'CONTACT_TYPE',			# Too general to leave in cloud.
	'CONTACT_DEVICE',
	'EmailAddress',
	'PhoneNumber',
	'WideAreaAccess',		# Node to WAN directory.
	'WideAreaLookup',
	'WideAreaRedirect',
	'WideAreaAssignment',
	'YourWideArea',
	'RelayLookup',			# Node to WAN relay.
	'RelayRedirect',
	'RelayAssignment',
	'YourRelay',
	'CloseRelay',
	'FOH_HOST',
]

# Contact details.
CONTACT_TYPE = ar.Enumeration(PERSONAL=0, BUSINESS=1, HOME=2, OTHER=3)
CONTACT_DEVICE = ar.Enumeration(MOBILE=0, FIXED_LINE=1)

class EmailAddress(object):
	def __init__(self, email_type=None, email_address=None):
		self.email_type = email_type
		self.email_address = email_address

	def __str__(self):
		s = CONTACT_TYPE.to_name(self.email_type)
		return f'[{s}] {self.email_address}'

class PhoneNumber(object):
	def __init__(self, phone_type=None, phone_device=None, phone_number=None):
		self.phone_type = phone_type
		self.phone_device = phone_device
		self.phone_number = phone_number

	def __str__(self):
		t = CONTACT_TYPE.to_name(self.phone_type)
		d = CONTACT_DEVICE.to_name(self.phone_device)
		return f'[{t}/{d}] {self.phone_number}'

#
#
CONTACT_SCHEMA = {
	"email_type": CONTACT_TYPE,
	"email_address": ar.Unicode(),
	"phone_type": CONTACT_TYPE,
	"phone_device": CONTACT_DEVICE,
	"phone_number": ar.Unicode(),
}

ar.bind(EmailAddress, object_schema=CONTACT_SCHEMA)
ar.bind(PhoneNumber, object_schema=CONTACT_SCHEMA)

# Protocol between a connecting directory and
# a cloud directory.
class WideAreaAccess(object):
	def __init__(self, access_ipp=None, encrypted=False, access_token=None, account_id=None, directory_id=None,
			product_name=None, product_instance=None):
		self.access_ipp = access_ipp or HostPort()	# Manifestly required.
		self.encrypted = encrypted
		self.access_token = access_token
		self.account_id = account_id				# Speed up processing.
		self.directory_id = directory_id
		self.product_name = product_name
		self.product_instance = product_instance

	def __str__(self):
		t = InstanceOfProduct.to_name(self.product_instance)
		return f'{self.product_name}/{t}'

class WideAreaLookup(object):
	def __init__(self, account_id=None, directory_id=None,
			access_token=None,
			product_name=None, product_instance=None):
		self.account_id = account_id
		self.directory_id = directory_id
		self.access_token = access_token
		self.product_name = product_name
		self.product_instance = product_instance

class WideAreaRedirect(object):
	def __init__(self, redirect_ipp=None, directory_id=None, assignment_token=None, encrypted=False):
		self.redirect_ipp = redirect_ipp or HostPort()
		self.directory_id = directory_id
		self.assignment_token = assignment_token
		self.encrypted = encrypted

class WideAreaAssignment(object):
	def __init__(self, directory_id=None, assignment_token=None):
		self.directory_id = directory_id
		self.assignment_token = assignment_token

class YourWideArea(object):
	def __init__(self, address=None):
		self.address = address

WIDE_AREA_SCHEMA = {
	"access_ipp": ar.UserDefined(HostPort),
	"access_token": ar.UUID(),
	"account_id": ar.UUID(),
	"directory_id": ar.UUID(),
	"redirect_ipp": ar.UserDefined(HostPort),
	"assignment_token": ar.UUID(),
	"address": ar.Address(),
	"product_name": ar.Unicode(),
	"product_instance": InstanceOfProduct,
	"encrypted": ar.Boolean(),
}

ar.bind(WideAreaAccess, object_schema=WIDE_AREA_SCHEMA)
ar.bind(WideAreaLookup, object_schema=WIDE_AREA_SCHEMA)
ar.bind(WideAreaRedirect, object_schema=WIDE_AREA_SCHEMA)
ar.bind(WideAreaAssignment, object_schema=WIDE_AREA_SCHEMA)
ar.bind(YourWideArea, object_schema=WIDE_AREA_SCHEMA)

# Protocol between a routing/looping subscriber and
# a publisher.
class RelayLookup(object):
	def __init__(self, relay_id=None, directory_id=None):
		self.relay_id = relay_id
		self.directory_id = directory_id

class RelayRedirect(object):
	def __init__(self, redirect_ipp=None, relay_id=None, assignment_token=None, encrypted=None):
		self.redirect_ipp = redirect_ipp or HostPort()
		self.relay_id = relay_id
		self.assignment_token = assignment_token
		self.encrypted = encrypted

class RelayAssignment(object):
	def __init__(self, relay_id=None, assignment_token=None):
		self.relay_id = relay_id
		self.assignment_token = assignment_token

class YourRelay(object):
	def __init__(self, address=None):
		self.address = address

class CloseRelay(object):
	def __init__(self, redirect=None):
		self.redirect = redirect or RelayRedirect()

RELAY_SCHEMA = {
	"relay_id": ar.UUID(),
	"directory_id": ar.UUID(),
	"redirect_ipp": ar.UserDefined(HostPort),
	"assignment_token": ar.UUID(),
	"address": ar.Address(),
	"account_id": ar.UUID(),
	"encrypted": ar.Boolean(),
}

ar.bind(RelayLookup, object_schema=RELAY_SCHEMA)
ar.bind(RelayRedirect, object_schema=RELAY_SCHEMA)
ar.bind(RelayAssignment, object_schema=RELAY_SCHEMA)
ar.bind(YourRelay, object_schema=RELAY_SCHEMA)

#
#
class CloseRelay(object):
	def __init__(self, redirect=None):
		self.redirect = redirect or RelayRedirect()

CLOSE_SCHEMA = {
	"redirect": ar.UserDefined(RelayRedirect),
}

ar.bind(CloseRelay, object_schema=CLOSE_SCHEMA)

FOH_HOST = 'ansar-mx.net'





# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
'''Ddclaration of the Group role, i.e. its settings.

Needed for sharing of group role across ansar-group and ansar (network, etc).
'''

import ansar.create as ar
from .socketry import *
from .directory_if import *

__all__ = [
	'GroupSettings',
]

class GroupSettings(object):
	def __init__(self, connect_above=None, roles=None, main_role=None, forwarding=False,
		show_scopes=None, connect_scope=None, connect_file=None, encrypted=None):
		self.connect_above = connect_above or HostPort()
		self.roles = roles
		self.main_role = main_role
		self.forwarding = forwarding
		self.show_scopes = show_scopes
		self.connect_scope = connect_scope
		self.connect_file = connect_file
		self.encrypted = encrypted

GROUP_SETTINGS_SCHEMA = {
	'connect_above': ar.Any(),
	'roles': ar.Unicode(),
	'main_role': ar.Unicode(),
	'forwarding': ar.Boolean(),
	'show_scopes': ar.Boolean(),
	'connect_scope': ScopeOfService,
	'connect_file': ar.Unicode(),
	'encrypted': ar.Boolean(),
}

ar.bind(GroupSettings, object_schema=GROUP_SETTINGS_SCHEMA)



# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
__docformat__ = 'restructuredtext'

import ansar.create as ar
from .socketry import *
from .transporting import *

__all__ = [
	'RETRY_LOCAL',
	'RETRY_PRIVATE',
	'RETRY_PUBLIC',
	'ip_retry',
]

# Reasonable intervals between connection attempts for
# the different scopes of address.
RETRY_LOCAL = ar.RetryIntervals(first_steps=[2.0, 4.0], regular_steps=8.0, step_limit=None, randomized=0.25, truncated=0.5)
RETRY_PRIVATE = ar.RetryIntervals(first_steps=[4.0, 8.0], regular_steps=16.0, step_limit=None, randomized=0.25, truncated=0.5)
RETRY_PUBLIC = ar.RetryIntervals(first_steps=[8.0, 16.0, 32.0], regular_steps=64.0, step_limit=None, randomized=0.25, truncated=0.5)

def ip_retry(s):
	if s == ScopeOfIP.OTHER:	# Not a dotted IP.
		return RETRY_PUBLIC

	if s == ScopeOfIP.LOCAL:		# Local - 127.
		return RETRY_LOCAL
	elif s == ScopeOfIP.PRIVATE:	# Private - 192.168.
		return RETRY_PRIVATE

	return RETRY_PUBLIC				# Domains?



# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
__docformat__ = 'restructuredtext'

import queue as sq
import threading as thr
import errno
import socket
import select
import re
import uuid
import nacl.utils
from nacl.public import PrivateKey, PublicKey, Box
import ansar.create as ar
from .transporting_if import ts

from copy import copy

__all__ = [
	'HostPort',
	'LocalPort',
	'ScopeOfIP',
	'local_private_public',
	'Blob',
	'CreateFrame',
	'ListenForStream',
	'ConnectStream',
	'StopListening',
	'Listening',
	'Accepted',
	'Connected',
	'NotListening',
	'NotAccepted',
	'NotConnected',
	'Close',
	'Closed',
	'Abandoned',
	'SocketSelect',
	'SocketChannel',
]

#
#
LOCAL_HOST = '127.0.0.1'

class HostPort(object):
	"""Combination of an IP address or name, and a port number.

	:param host: IP address or name
	:type host: str
	:param port: network port
	:type port: int
	"""
	def __init__(self, host=None, port=None):
		self.host = host
		self.port = port
	
	def __str__(self):
		return f'{self.host}:{self.port}'

	def inet(self):
		return (self.host, self.port)

class LocalPort(HostPort):
	def __init__(self, port=None):
		HostPort.__init__(self, LOCAL_HOST, port)

HOST_PORT_SCHEMA = {
	'host': str,
	'port': int,
}

ar.bind(HostPort, object_schema=HOST_PORT_SCHEMA)
ar.bind(LocalPort, object_schema=HOST_PORT_SCHEMA)

#
#
DOTTED_IP = re.compile(r'(\d+)\.(\d+)\.(\d+)\.(\d+)')
ScopeOfIP = ar.Enumeration(OTHER=0, LOCAL=1, PRIVATE=2, PUBLIC=3)

def local_private_public(ip):
	m = DOTTED_IP.match(ip)
	if m is None:
		return ScopeOfIP.OTHER
	# Have complete verification of dotted layout
	b0 = int(m.groups()[0])
	b1 = int(m.groups()[1])

	# Not dotted -------- None
	# 127.x.x.x --------- 0, localhost
	# 10.x.x.x ---------- 1, private
	# 192.168.x.x ------- 1, private
	# 172.[16-31].x.x --- 1, private
	# else -------------- 2, public

	if b0 == 127:
		return ScopeOfIP.LOCAL
	elif b0 == 10:
		return ScopeOfIP.PRIVATE
	elif b0 == 192 and b1 == 168:
		return ScopeOfIP.PRIVATE
	elif b0 == 172 and (b1 > 15 and b1 < 32):
		return ScopeOfIP.PRIVATE
	return ScopeOfIP.PUBLIC

#
#
class Blob(object):
    def __init__(self, block=None):
        self.block = block

ar.bind(Blob, object_schema={'block': ar.Block()})

#
#
class CreateFrame(object):
	"""Capture values needed for async object creation.

	:param object_type: type to be created
	:type object_type: function or Point-based class
	:param args: positional parameters
	:type args: tuple
	:param kw: named parameters
	:type kw: dict
	"""
	def __init__(self, object_type, *args, **kw):
		self.object_type = object_type
		self.args = args
		self.kw = kw

# Control messages sent to the sockets thread
# via the control channel.
class ListenForStream(object):
	def __init__(self, requested_ipp=None, create_session=None, upgrade=None, encrypted=False):
		self.requested_ipp = requested_ipp or HostPort()
		self.create_session = create_session
		self.upgrade = upgrade
		self.encrypted = encrypted

class ConnectStream(object):
	def __init__(self, requested_ipp=None, create_session=None, upgrade=None, encrypted=False):
		self.requested_ipp = requested_ipp or HostPort()
		self.create_session = create_session
		self.upgrade = upgrade
		self.encrypted = encrypted

class StopListening(object):
	def __init__(self, listening_ipp=None):
		self.listening_ipp = listening_ipp or HostPort()

# Update messages from sockets thread to app.
class Listening(object):
	"""Session notification, server presence established.

	:param requested_ipp: IP and port to listen at
	:type requested_ipp: HostPort
	:param listening_ipp: established IP and port
	:type listening_ipp: HostPort
	:param encrypted: is the client encrypting
	:type encrypted: bool
	"""
	def __init__(self, requested_ipp=None, listening_ipp=None, encrypted=False):
		self.requested_ipp = requested_ipp or HostPort()
		self.listening_ipp = listening_ipp or HostPort()
		self.encrypted = encrypted

class Accepted(object):
	"""Session notification, transport to client established.

	:param requested_ipp: IP and port listening at
	:type requested_ipp: HostPort
	:param accepted_ipp: local IP and port
	:type accepted_ipp: HostPort
	:param remote_address: address of SocketProxy
	:type remote_address: async address
	:param opened_at: moment of acceptance
	:type opened_at: datetime
	"""
	def __init__(self, listening_ipp=None, accepted_ipp=None, remote_address=None, opened_at=None):
		self.listening_ipp = listening_ipp or HostPort()
		self.accepted_ipp = accepted_ipp or HostPort()
		self.remote_address = remote_address
		self.opened_at = opened_at

class Connected(object):
	"""Session notification, transport to server established.

	:param requested_ipp: IP and port to connect to
	:type requested_ipp: HostPort
	:param connected_ipp: local IP and port
	:type connected_ipp: HostPort
	:param remote_address: address of SocketProxy
	:type remote_address: async address
	:param opened_at: moment of connection
	:type opened_at: datetime
	"""
	def __init__(self, requested_ipp=None, connected_ipp=None, remote_address=None, opened_at=None):
		self.requested_ipp = requested_ipp or HostPort()
		self.connected_ipp = connected_ipp or HostPort()
		self.remote_address = remote_address
		self.opened_at = opened_at

class NotListening(ar.Faulted):
	"""Session notification, server not established.

	:param requested_ipp: IP and port to listen at
	:type requested_ipp: HostPort
	:param error_code: platform error number
	:type error_code: int
	:param error_text: platform error message
	:type error_text: str
	"""
	def __init__(self, requested_ipp=None, error_code=0, error_text=None):
		ar.Faulted.__init__(self, f'cannot listen at "{requested_ipp}"', error_text, exit_code=error_code)
		self.requested_ipp = requested_ipp or HostPort()

class NotAccepted(ar.Faulted):
	"""Session notification, transport to client not established.

	:param listening_ipp: IP and port listening at
	:type listening_ipp: HostPort
	:param error_code: platform error number
	:type error_code: int
	:param error_text: platform error message
	:type error_text: str
	"""
	def __init__(self, listening_ipp=None, error_code=0, error_text=None):
		ar.Faulted.__init__(self, f'cannot accept at "{listening_ipp}"', error_text, exit_code=error_code)
		self.listening_ipp = listening_ipp or HostPort()

class NotConnected(ar.Faulted):
	"""Session notification, transport to server established.

	:param requested_ipp: IP and port to connect to
	:type requested_ipp: HostPort
	:param error_code: platform error number
	:type error_code: int
	:param error_text: platform error message
	:type error_text: str
	"""
	def __init__(self, requested_ipp=None, error_code=0, error_text=None):
		ar.Faulted.__init__(self, f'cannot connect to "{requested_ipp}"', error_text, exit_code=error_code)
		self.requested_ipp = requested_ipp or HostPort()

CONTROL_SCHEMA = {
	'requested_ipp': ar.UserDefined(HostPort),
	'controller_address': ar.Address(),
	'remote_address': ar.Address(),
	'opened_at': ar.WorldTime(),
	'upgrade': ar.Type(),
	'create_session': ar.Type(),
	'connected_ipp': ar.UserDefined(HostPort),
	'listening_ipp': ar.UserDefined(HostPort),
	'accepted_ipp': ar.UserDefined(HostPort),
	'condition': ar.Unicode(),
	'explanation': ar.Unicode(),
	'error_code': ar.Integer8(),
	'error_text': ar.Unicode(),
	'exit_code': ar.Integer8(),
	'encrypted': ar.Boolean(),
}

ar.bind(ListenForStream, object_schema=CONTROL_SCHEMA)
ar.bind(ConnectStream, object_schema=CONTROL_SCHEMA)
ar.bind(StopListening, object_schema=CONTROL_SCHEMA)
ar.bind(Listening, object_schema=CONTROL_SCHEMA, copy_before_sending=False)
ar.bind(Accepted, object_schema=CONTROL_SCHEMA)
ar.bind(Connected, object_schema=CONTROL_SCHEMA)
ar.bind(NotListening, object_schema=CONTROL_SCHEMA)
ar.bind(NotAccepted, object_schema=CONTROL_SCHEMA)
ar.bind(NotConnected, object_schema=CONTROL_SCHEMA)

# Session termination messages. Handshake between app
# and sockets thread to cleanly terminate a connection.
class Close(object):
	"""Session control, terminate the messaging transport.

	:param value: completion value for the session
	:type value: any
	"""
	def __init__(self, value=None):
		self.value = value

class Closed(ar.Faulted):
	"""Session notification, local termination of the messaging transport.

	:param value: completion value for the session
	:type value: any
	:param reason: short description
	:type reason: str
	:param opened_ipp: local IP address and port number
	:type opened_ipp: HostPort
	:param opened_at: moment of termination
	:type opened_at: datetime
	"""
	def __init__(self, value=None, reason=None, opened_ipp=None, opened_at=None):
		ar.Faulted.__init__(self, f'closed {opened_ipp}', reason)
		self.value = value
		self.opened_ipp = opened_ipp or HostPort()
		self.opened_at = opened_at

class Abandoned(ar.Faulted):
	"""Session notification, remote termination of the messaging transport.

	:param opened_ipp: local IP address and port number
	:type opened_ipp: HostPort
	:param opened_at: moment of termination
	:type opened_at: datetime
	"""
	def __init__(self, opened_ipp=None, opened_at=None):
		ar.Faulted.__init__(self, f'abandoned by {opened_ipp}')
		self.opened_ipp = opened_ipp or HostPort()
		self.opened_at = opened_at

ENDING_SCHEMA = {
	'value': ar.Any,
	'opened_ipp': ar.UserDefined(HostPort),
	'opened_at': ar.WorldTime(),
	'condition': ar.Unicode(),
	'explanation': ar.Unicode(),
	'error_code': ar.Integer8(),
	'error_text': ar.Unicode(),
	'exit_code': ar.Integer8(),
}

ar.bind(Close, object_schema=ENDING_SCHEMA, copy_before_sending=False)
ar.bind(Closed, object_schema=ENDING_SCHEMA, copy_before_sending=False)
ar.bind(Abandoned, object_schema=ENDING_SCHEMA, copy_before_sending=False)

#
#
class Shutdown(object):
	def __init__(self, s=None, value=False):
		self.s = s
		self.value = value

class Bump(object):
	def __init__(self, s=None):
		self.s = s

QUEUE_SCHEMA = {
	's': ar.Any,
	'value': ar.Any,
	'stream': ar.Any,
}

ar.bind(Shutdown, object_schema=QUEUE_SCHEMA, copy_before_sending=False)
ar.bind(Bump, object_schema=QUEUE_SCHEMA, copy_before_sending=False)

# Classes representing open sockets for one reason or another;
# - ControlChannel.... accepted end of backdoor into sockets loop.
# - TcpServer ........ an active listen
# - TcpClient ........ an active connect
# - TcpStream ........ established transport, child of listen or connect

class ControlChannel(object):
	def __init__(self, s):
		self.s = s

class TcpServer(object):
	def __init__(self, s, request, listening, controller_address, upgrade):
		self.s = s
		self.request = request
		self.listening = listening
		self.controller_address = controller_address
		self.upgrade = upgrade

class TcpClient(object):
	def __init__(self, s, request, connected, controller_address, upgrade, encrypted):
		self.s = s
		self.request = request
		self.connected = connected
		self.controller_address = controller_address
		self.upgrade = upgrade
		self.encrypted = encrypted

# Underlying network constraints.
#
TCP_RECV = 4096
TCP_SEND = 4096
UDP_RECV = 4096
UDP_SEND = 4096

# Security/reliability behaviours.
#
NUMBER_OF_DIGITS = (7 * 3) + 2
GIANT_FRAME = 1048576

#
#
class Header(object):
	def __init__(self, to_address=None, return_address=None, tunnel=False):
		self.to_address = to_address
		self.return_address = return_address
		self.tunnel = tunnel

HEADER_SCHEMA = {
	"to_address": ar.TargetAddress(),
	"return_address": ar.Address(),
	"tunnel": ar.Boolean(),
}

ar.bind(Header, object_schema=HEADER_SCHEMA)

HEADING = ar.UserDefined(Header)
SPACE = ar.VectorOf(ar.Address())

#
#
class Relay(object):
	def __init__(self, block=None, space=None):
		self.block = block
		self.space = space

RELAY_SCHEMA = {
	"block": ar.Block(),
	"space": ar.VectorOf(ar.Address()),
}

ar.bind(Relay, object_schema=RELAY_SCHEMA)

class StreamingIn(object):
	def __init__(self):
		self.analysis_state = 1
		self.size_byte = bytearray()
		self.size_len = []
		self.frame_size = 0
		self.frame_byte = bytearray()

		def s1(c):
			if c in b'0123456789,':
				nd = len(self.size_byte)
				if nd < NUMBER_OF_DIGITS:
					self.size_byte.append(c)
					return 1
				raise OverflowError(f'unlikely frame size with {nd} digits')
			elif c == 10:	# ord('\n')
				a = self.size_byte.split(b',')
				if len(a) != 3:
					raise ValueError(f'unexpected dimension')
				for b in a:
					if not b or not b.isdigit():
						raise ValueError(f'mangled frame dimensions')
				s0 = int(a[0])
				s1 = int(a[1])
				s2 = int(a[2])
				self.size_len = [s0, s1, s2]
				if s0 > s2 or (s0 + s1) > s2:
					raise ValueError(f'unlikely frame offsets')
				self.jump_size = s2
				if self.jump_size > GIANT_FRAME:
					raise OverflowError(f'oversize frame of {self.jump_size} bytes')
				elif self.jump_size == 0:
					return 3
				return 2
			raise ValueError(f'frame with unexpected {c} in digits')

		def s2(c):
			self.frame_byte.append(c)
			self.jump_size -= 1
			if self.jump_size == 0:
				return 3
			return 2

		def s3(c):
			if c == 10:
				return 0
			raise ValueError(f'unexpected {c} at end-of-frame')

		self.shift = {
			1: s1,
			2: s2,
			3: s3,
		}

	def recover_frame(self, received):
		for c in received:
			next = self.shift[self.analysis_state](c)
			if next:
				self.analysis_state = next
				continue

			# Completed frame.
			f = bytes(self.frame_byte)
			if self.key_box:
				f = self.key_box.decrypt(f)

			# Breakout parts and yield.
			n0 = self.size_len[0]
			n1 = self.size_len[1]
			b2 = n0 + n1
			yield f[0:n0], f[n0:b2], f[b2:]

			# Restart.
			self.analysis_state = 1
			self.size_byte = bytearray()
			self.size_len = []
			self.frame_size = 0
			self.frame_byte = bytearray()


class StreamingOut(object):
	def __init__(self):
		self.pending = []		   # Messages not yet in the loop.
		self.lock = thr.RLock()	 # Safe sharing and empty detection.
		self.messages_to_encode = ar.deque()

	def put(self, m, t, r):
		try:
			self.lock.acquire()
			empty = len(self.pending) == 0
			t3 = (m, t, r)
			self.pending.append(t3)
		finally:
			self.lock.release()
		return empty

	def drain(self, a):
		try:
			self.lock.acquire()
			count = len(self.pending)
			a.extend(self.pending)
			self.pending = []
		finally:
			self.lock.release()
		return count

	def stream_frame(self, encoded_bytes, codec):
		while len(encoded_bytes) < TCP_SEND:
			if len(self.messages_to_encode) == 0:
				added = self.drain(self.messages_to_encode)
				if added == 0:
					break
			# The message and to-return addresses.
			m, t, r = self.messages_to_encode.popleft()
			f = False
			key_box = self.key_box

			# Types significant to streaming.
			# Be nice to move DH detection elsewhere.
			if isinstance(m, Blob):
				f = True
			elif isinstance(m, Diffie):
				self.private_key = PrivateKey.generate()
				public_bytes = self.private_key.public_key.encode()
				m.public_key = bytearray(public_bytes)
			elif isinstance(m, Hellman):
				self.private_key = PrivateKey.generate()
				shared_bytes = bytes(m.public_key)
				shared_key = PublicKey(shared_bytes)
				self.key_box = Box(self.private_key, shared_key)
				public_bytes = self.private_key.public_key.encode()
				m.public_key = bytearray(public_bytes)

			# Bring the parts together.
			# 1. Header
			h = Header(t, r, f)
			e = codec.encode(h, HEADING)
			b0 = e.encode('utf-8')
			n0 = len(b0)

			# 2. Message body - 1 of following 3.
			if f:
				b1 = m.block
				n1 = len(b1)
				s = codec.encode([], SPACE)
			elif isinstance(m, Relay):
				b1 = m.block
				n1 = len(b1)
				s = codec.encode(m.space, SPACE)
			else:
				space = []
				e = codec.encode(m, ar.Any(), space=space)
				b1 = e.encode('utf-8')
				n1 = len(b1)
				s = codec.encode(space, SPACE)

			# 3. Mutated addresses.
			b2 = s.encode('utf-8')

			# Combine into 1 and optionally encrypt.
			b0 += b1
			b0 += b2
			if key_box:
				b0 = key_box.encrypt(b0)
			n3 = len(b0)

			# Put frame on the stream.
			n = f'{n0},{n1},{n3}'
			encoded_bytes += n.encode('ascii')
			encoded_bytes += b'\n'
			encoded_bytes += b0
			encoded_bytes += b'\n'
		return len(encoded_bytes)

class TcpStream(StreamingIn, StreamingOut):
	def __init__(self, parent, controller_address, upgrade, opened):
		StreamingIn.__init__(self)
		StreamingOut.__init__(self)
		self.parent = parent
		self.controller_address = controller_address
		self.remote_address = None
		self.upgrade = upgrade
		self.opened = opened
		self.closing = False
		self.value = None
		self.codec = None
		self.encoded_bytes = bytearray()
		self.diffie_hellman = None
		self.private_key = None
		self.key_box = None

	def routing(self, return_proxy, local_termination, remote_address):
		# Define addresses for message forwarding.
		# return_proxy ........ address that response should go back to.
		# local_termination ... address of default target, actor or session.
		# remote_address ...... source address of connection updates, session or proxy.
		self.codec = ar.CodecJson(return_proxy=return_proxy, local_termination=local_termination)
		self.remote_address = remote_address

	def receive_and_route(self, received, sockets):
		for h, b_, a in self.recover_frame(received):
			s = h.decode('utf-8')
			header, v = self.codec.decode(s, HEADING)
			if v is not None:
				raise ValueError(f'header with unexpected versioning "{v}"')
			s = a.decode('utf-8')
			space, v = self.codec.decode(s, SPACE)

			to_address = header.to_address

			if header.tunnel:
				sockets.forward(Blob(b_), to_address, header.return_address)
				continue
			elif len(header.to_address) > 1:
				sockets.forward(Relay(b_, space), to_address, header.return_address)
				continue

			s = b_.decode('utf-8')
			body, v = self.codec.decode(s, ar.Any(), space=space)
			if v is not None:
				if not self.upgrade:
					raise ValueError(f'body version "{v}" and no upgrade')
				body = self.upgrade(body, v)

			if isinstance(body, Diffie):
				sockets.send(Hellman(body.public_key), self.remote_address)
				if self.diffie_hellman:
					h = self.diffie_hellman[0]
					sockets.forward(h[0], h[1], h[2])
			elif isinstance(body, Hellman):
				public_bytes = bytes(body.public_key)
				public_key = PublicKey(public_bytes)
				self.key_box = Box(self.private_key, public_key)
				if self.diffie_hellman:
					h = self.diffie_hellman[0]
					sockets.forward(h[0], h[1], h[2])
			else:
				sockets.forward(body, to_address, header.return_address)

	def send_a_block(self, s):
		t = self.stream_frame(self.encoded_bytes, self.codec)
		if t == 0:
			return False
		n = t if t <= TCP_SEND else TCP_SEND
		chunk = self.encoded_bytes[:n]
		n = s.send(chunk)
		if n:
			self.encoded_bytes = self.encoded_bytes[n:]
			return True
		return False

#
#
class INITIAL: pass
class PENDING: pass
class NORMAL: pass

class SocketProxy(ar.Point, ar.StateMachine):
	"""Local representation of an object at remote end of a network connection.

	:param s: associated network connection
	:type s: socket descriptor
	:param channel: async socket loop
	:type channel: internal
	:param stream: associated buffering
	:type stream: internal
	"""
	def __init__(self, s, channel, stream):
		ar.Point.__init__(self)
		ar.StateMachine.__init__(self, INITIAL)
		self.s = s
		self.channel = channel
		self.stream = stream

SOCKET_DOWN = (errno.ECONNRESET, errno.EHOSTDOWN, errno.ENETDOWN, errno.ENETRESET)

def SocketProxy_INITIAL_Start(self, message):
	return NORMAL

def SocketProxy_NORMAL_Unknown(self, message):
	empty = self.stream.put(message, self.to_address, self.return_address)
	if empty:
		self.channel.send(Bump(self.s), self.address)
	return NORMAL

def SocketProxy_NORMAL_Close(self, message):
	self.channel.send(Shutdown(self.s, message.value), self.address)
	self.complete()

def SocketProxy_NORMAL_Stop(self, message):
	self.complete()

TCP_PROXY_DISPATCH = {
	INITIAL: (
		(ar.Start,),
		()
	),
	NORMAL: (
		(ar.Unknown, Close, ar.Stop),
		()
	),
}

ar.bind(SocketProxy, TCP_PROXY_DISPATCH)

#
#
class Diffie(object):
	def __init__(self, public_key=None):
		self.public_key = public_key

class Hellman(object):
	def __init__(self, public_key=None):
		self.public_key = public_key

DH_SCHEMA = {
	'public_key': ar.Block(),
}

ar.bind(Diffie, object_schema=DH_SCHEMA, copy_before_sending=False)
ar.bind(Hellman, object_schema=DH_SCHEMA, copy_before_sending=False)



# Signals from the network represented
# as distinct classes - for dispatching.
class ReceiveBlock: pass
class ReadyToSend: pass
class BrokenTransport: pass

# CONTROL CHANNEL
# First two functions are for handling the 1-byte events
# coming across the control socket.
def ControlChannel_ReceiveBlock(self, control, s):
	s.recv(1)					   # Consume the bump.
	mr = self.pending.get()	# Message and account.

	# This second jump is to simulate the common handling of control
	# channel events and select events.
	c = type(mr[0])
	try:
		f = select_table[(ControlChannel, c)]
	except KeyError:
		self.fault('Unknown signal received at control channel (%s).' % (c.__name__,))
		return
	f(self, control, mr)

def ControlChannel_BrokenTransport(self, control, s):
	self.fault('The control channel to the selector is broken.')
	self.clear(s, ControlChannel)

# The rest of them handle the simulated receive of the
# actual message.

def ControlChannel_ListenForStream(self, control, mr):
	m, r = mr
	requested_ipp = m.requested_ipp
	try:
		server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	except socket.error as e:
		self.send(NotListening(requested_ipp, e.errno, str(e)), r)
		return

	def server_not_listening(e):
		server.close()
		self.send(NotListening(requested_ipp, e.errno, str(e)), r)

	try:
		server.setblocking(False)
		server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		server.bind(requested_ipp.inet())
		server.listen(5)
	except socket.herror as e:
		server_not_listening(e)
		return
	except socket.gaierror as e:
		server_not_listening(e)
		return
	except socket.error as e:
		server_not_listening(e)
		return
	except OverflowError as e:
		server.close()
		self.send(NotListening(requested_ipp, 0, str(e)), r)
		return

	hap = server.getsockname()
	listening_ipp=HostPort(hap[0], hap[1])

	if m.encrypted:
		self.trace(f'Listening (encrypted) on "{listening_ipp}", requested "{requested_ipp}"')
	else:
		self.trace(f'Listening on "{listening_ipp}", requested "{requested_ipp}"')

	listening = Listening(requested_ipp=requested_ipp,
		listening_ipp=listening_ipp,
		encrypted=m.encrypted)

	self.networking[server] = TcpServer(server, m, listening, r, m.upgrade)
	self.receiving.append(server)
	self.faulting.append(server)

	self.send(listening, r)

def no_ending(value, parent, address):
	pass

def close_ending(proxy):
	def ending(value, parent, address):
		ar.send_a_message(Close(value), proxy, address)
	return ending

def open_stream(self, parent, s, opened):
	controller_address = parent.controller_address

	stream = TcpStream(parent, controller_address, parent.upgrade, opened)
	proxy_address = self.create(SocketProxy, s, self.channel, stream, object_ending=no_ending)

	cs = parent.request.create_session
	if cs:
		# Create the ending function that swaps the Completed message to the parent for a
		# Close message to the proxy.

		ending = close_ending(proxy_address)
		session_address = self.create(cs.object_type, *cs.args,
			controller_address=controller_address, remote_address=proxy_address,
			object_ending=ending,
			**cs.kw)
		stream.routing(proxy_address, session_address, session_address)
	else:
		stream.routing(proxy_address, controller_address, proxy_address)

	self.networking[s] = stream
	return stream, proxy_address

def ControlChannel_ConnectStream(self, control, mr):
	m, r = mr
	requested_ipp = m.requested_ipp
	try:
		client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		client.setblocking(False)
	except socket.error as e:
		self.send(NotConnected(requested_ipp, e.errno, str(e)), r)
		return

	def client_not_connected(e):
		client.close()
		self.send(NotConnected(requested_ipp, e.errno, str(e)), r)

	try:
		e = client.connect_ex(requested_ipp.inet())
		if e:
			# Connect request cannot complete. Check for codes indicating
			# async issue. If not it's a real error.
			if e not in (errno.EINPROGRESS, errno.EWOULDBLOCK, errno.EAGAIN):
				client.close()
				self.send(NotConnected(requested_ipp, e, 'Connect incomplete and no pending indication.'), r)
				return

			# Build a transient "session" that just exists to catch
			# an initial, either send or fault (a receive is treated
			# as an error). True session is constructed on receiving
			# a "normal" send event.
			pending = TcpClient(client, m, None, r, m.upgrade, m.encrypted)

			self.networking[client] = pending
			self.receiving.append(client)
			self.sending.append(client)
			self.faulting.append(client)
			return

	except socket.herror as e:
		client_not_connected(e)
		return
	except socket.gaierror as e:
		client_not_connected(e)
		return
	except socket.error as e:
		client_not_connected(e)
		return
	except OverflowError as e:
		client.close()
		self.send(NotConnected(requested_ipp, 0, str(e)), r)
		return

	hap = client.getsockname()
	connected_ipp = HostPort(hap[0], hap[1])

	connected = Connected(requested_ipp=requested_ipp,
		connected_ipp=connected_ipp,
		opened_at=ar.world_now())

	parent = TcpClient(client, m, connected, r, m.upgrade, m.encrypted)
	stream, proxy_address = open_stream(self, parent, client, connected)
	connected.remote_address = proxy_address

	self.networking[client] = stream
	self.receiving.append(client)
	self.sending.append(client)
	self.faulting.append(client)

	if m.encrypted:
		self.trace(f'Connected (encrypted) to "{requested_ipp}", at local address "{connected_ipp}"')
		not_connected = NotConnected(requested_ipp, None, None)
		stream.diffie_hellman = (
			(connected, r, stream.remote_address),
			(not_connected, r))
		# Start the exchange. Public key filled out during
		# streaming.
		self.send(Diffie(), proxy_address)
		return
	self.trace(f'Connected to "{requested_ipp}", at local address ""{connected_ipp}"')

	self.forward(connected, r, stream.remote_address)

def ControlChannel_StopListening(self, control, mr):
	m, r = mr
	listening_ipp = m.listening_ipp
	def server(t):
		if not isinstance(t, TcpServer):
			return False
		h = t.listening.listening_ipp.host == listening_ipp.host
		p = t.listening.listening_ipp.port == listening_ipp.port
		return h and p

	# Find server belonging to sender
	# and clear from engine.
	sockets = [k for k, v in self.networking.items() if server(v)]
	if len(sockets) == 1:
		self.clear(sockets[0], TcpServer)
		text = 'stopped "%s"(%d)' % (listening_ipp.host, listening_ipp.port)
	else:
		text = 'not listening to "%s"(%d)' % (listening_ipp.host, listening_ipp.port)
	self.send(NotListening(listening_ipp, 0, text), r)

def ControlChannel_Stop(self, control, mr):
	m, r = mr
	def soc(p): # Server or client.
		return isinstance(p, (TcpServer, TcpClient))

	# Clear any servers and clients. Not
	# accepting or connecting any more.
	sockets = [k for k, v in self.networking.items() if soc(v)]
	for s in sockets:
		self.clear(s)

	# Only streams left. Except control channel. Sigh.
	# Kick off a proper teardown and wait until the handshaking
	# is done and there is nothing left to do.
	for k, v in self.networking.items():
		if isinstance(v, TcpStream):
			self.send(Close(), v.remote_address)

	self.running = False

def ControlChannel_Bump(self, control, mr):
	m, r = mr
	if m.s.fileno() < 0:
		# Catches the situation where the socket has been abandoned
		# by the remote and the notification to the proxy arrives behind
		# the bump.
		return
	try:
		self.sending.index(m.s)
		return
	except ValueError:
		pass
	self.sending.append(m.s)

def ControlChannel_Shutdown(self, control, mr):
	m, r = mr
	try:
		stream = self.networking[m.s]
	except KeyError:
		# Already cleared by Abandoned codepath.
		return
	stream.closing = True
	stream.value = m.value
	m.s.shutdown(socket.SHUT_RD)

# Dispatch of socket signals;
# - ReceiveBlock ...... there are bytes to recv
# - ReadyToSend ....... an opportunity to send
# - BrokenTransport ... error on socket
# and server/client/connection;
# - TcpServer ......... listen waiting to accept
# - TcpClient ......... partial connect
# - TcpStream ......... established connection

def TcpServer_ReceiveBlock(self, server, s):
	listening = server.listening
	try:
		accepted, hap = s.accept()
		accepted.setblocking(False)
	except socket.error as e:
		self.send(NotAccepted(listening.requested_ipp, e.errno, str(e)), server.controller_address)
		return

	opened_at = ar.world_now()
	stream, proxy_address = open_stream(self, server, accepted, None)
	self.receiving.append(accepted)
	self.sending.append(accepted)
	self.faulting.append(accepted)

	accepted_ipp = HostPort(hap[0], hap[1])

	accepted = Accepted(listening_ipp=listening.listening_ipp,
		accepted_ipp=accepted_ipp, remote_address=stream.remote_address,
		opened_at=opened_at)
	stream.opened = accepted

	if listening.encrypted:
		self.trace(f'Accepted (encrypted) "{accepted_ipp}", requested "{listening.listening_ipp}"')
		not_accepted = NotAccepted(listening.requested_ipp, None, None)
		stream.diffie_hellman = (
			(accepted, server.controller_address, stream.remote_address),
			(not_accepted, server.controller_address))
		return
	self.trace(f'Accepted "{accepted_ipp}", requested "{listening.listening_ipp}"')

	self.forward(accepted, server.controller_address, stream.remote_address)

def TcpServer_BrokenTransport(self, server, s):
	listening = server.listening
	self.send(NotListening(listening.listening_ipp, 0, "signaled by networking subsystem"), server.controller_address)
	self.clear(s, TcpServer)

# TCP CLIENT
# A placeholder for the eventual outbound stream.
def TcpClient_ReceiveBlock(self, selector, s):
	client = s
	# NOT NEEDED IN TcpStream_ReceiveBlock SO....
	#self.sending.remove(client)

	request = selector.request

	hap = client.getsockname()
	connected_ipp = HostPort(hap[0], hap[1])
	requested_ipp = request.requested_ipp

	# CANNOT BUILD A STREAM AND IMMEDIATELY TEAR IT DOWN ON AN EXCEPTION.
	# THIS WILL MAY CREATE A SESSION OBJECT WHEN THERE IS NO REMOTE AND MAY
	# NEVER BE. DO IT AFTER A SUCCESSFUL RECV().
	#connected = Connected(requested_ipp=request.requested_ipp,
	#	connected_ipp=HostPort(hap[0], hap[1]),
	#	opened_at=ar.world_now())
	#selector.connected = connected

	#stream, proxy_address = open_stream(self, selector, client, connected.opened_at)
	#connected.remote_address = proxy_address

	try:
		scrap = s.recv(TCP_RECV)

		# No exception. New stream.
		connected = Connected(requested_ipp=requested_ipp,
			connected_ipp=connected_ipp,
			opened_at=ar.world_now())

		selector.connected = connected
		stream, proxy_address = open_stream(self, selector, client, connected)
		connected.remote_address = proxy_address

		self.trace(f'Connected to "{requested_ipp}", at local address "{connected_ipp}"')

		if selector.encrypted:
			self.trace(f'Connected (encrypted) to "{requested_ipp}", at local address "{connected_ipp}"')
			not_connected = NotConnected(requested_ipp, None, None)
			stream.diffie_hellman = (
				(connected, stream.controller_address, stream.remote_address),
				(not_connected, selector.controller_address))
			self.send(Diffie(), proxy_address)
			return
		self.trace(f'Connected to "{requested_ipp}", at local address "{connected_ipp}"')

		self.forward(connected, stream.controller_address, stream.remote_address)

		if not scrap:
			# Immediate shutdown. Need to
			# generate the full set of messages.
			#self.clear(s, TcpStream)
			return

		try:
			stream.receive_and_route(scrap, self)
		except (ar.CodecFailed, OverflowError, ValueError) as e:
			value = ar.Faulted(condition='cannot stream inbound', explanation=str(e))
			self.warning(str(value))
			close_session(stream, value, s)
		return

	except socket.error as e:
		self.send(NotConnected(request.requested_ipp, e.errno, str(e)), selector.controller_address)
		self.clear(s, TcpClient)
		#self.send(NotConnected(request.requested_ipp, e.errno, str(e)), stream.controller_address)
		#self.send(ar.Stop(), stream.remote_address)
		#self.clear(s, TcpStream)
		return

def TcpClient_ReadyToSend(self, selector, s):
	client = s
	#self.sending.remove(client)

	request = selector.request
	hap = client.getsockname()
	connected_ipp = HostPort(hap[0], hap[1])
	requested_ipp = request.requested_ipp

	connected = Connected(requested_ipp=requested_ipp,
		connected_ipp=connected_ipp,
		opened_at=ar.world_now())
	selector.connected = connected

	stream, proxy_address = open_stream(self, selector, client, connected)
	connected.remote_address = proxy_address
	#receiving.append( client)
	#self.faulting.append( client)

	if selector.encrypted:
		self.trace(f'Connected (encrypted) to "{requested_ipp}", at local address "{connected_ipp}"')
		not_connected = NotConnected(requested_ipp, None, None)
		stream.diffie_hellman = (
			(connected, stream.controller_address, stream.remote_address),
			(not_connected, selector.controller_address))
		# Start the exchange of public keys.
		self.send(Diffie(), proxy_address)
		return
	self.trace(f'Connected to "{requested_ipp}", at local address "{connected_ipp}"')

	self.forward(connected, stream.controller_address, stream.remote_address)

def TcpClient_BrokenTransport(self, selector, s):
	text = 'fault on pending connect, unreachable, no service at that address or blocked'
	self.send(NotConnected(selector.requested_ipp, 0, text), selector.controller_address)
	self.clear(s, TcpClient)

def close_session(stream, value, s):
	stream.closing = True
	stream.value = value
	s.shutdown(socket.SHUT_RDWR)

def end_of_session(self, stream, s, reason=None):
	if isinstance(stream.opened, Connected):
		ipp = stream.opened.connected_ipp
	elif isinstance(stream.opened, Accepted):
		ipp = stream.opened.accepted_ipp
	else:
		ipp = None

	if stream.closing:
		c = Closed(value=stream.value,
			reason=reason,
			opened_ipp=ipp,
			opened_at=stream.opened.opened_at)
		self.forward(c, stream.controller_address, stream.remote_address)
	else:
		self.send(ar.Stop(), stream.remote_address)
		a = Abandoned(opened_ipp=ipp,
			opened_at=stream.opened.opened_at)
		self.forward(a, stream.controller_address, stream.remote_address)
	self.clear(s, TcpStream)

def TcpStream_ReadyToSend(self, stream, s):
	try:
		if stream.send_a_block(s):
			return
	except (ar.CodecFailed, OverflowError, ValueError) as e:
		value = ar.Faulted(condition='cannot stream outbound', explanation=str(e))
		self.warning(str(value))
		close_session(stream, value, s)
		return

	try:
		self.sending.remove(s)
	except KeyError:
		pass

# A network transport for the purpose of exchanging
# messages between machines.

def TcpStream_ReceiveBlock(self, stream, s):
	try:
		scrap = s.recv(TCP_RECV)
		if not scrap:
			end_of_session(self, stream, s, 'empty socket')
			return

		try:
			stream.receive_and_route(scrap, self)
		except (ar.CodecFailed, OverflowError, ValueError) as e:
			value = ar.Faulted(condition='cannot stream inbound', explanation=str(e))
			self.warning(str(value))
			close_session(stream, value, s)
		return

	except socket.error as e:
		if e.errno == errno.ECONNREFUSED:
			self.fault('Connection refused')
		elif e.errno not in SOCKET_DOWN:
			self.fault('Socket termination [%d] %s' % (e.errno, e.strerror))
		end_of_session(self, stream, s, reason=e.strerror)
		return

def TcpStream_BrokenTransport(self, selector, s):
	end_of_session(self, selector, s, reason='broken socket')

#
#
def control_channel():
	server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	server.setblocking(False)
	server.bind(("127.0.0.1", 0))
	server.listen(1)

	server_address = server.getsockname()

	client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
	client.setblocking(False)
	e = client.connect_ex(server_address)

	readable, writable, exceptional = select.select([server], [], [server])
	if not readable:
		client.close()
		server.close()
		raise RuntimeError('Forming control channel, select has not received connect notification.')

	accepted, client_address = server.accept()
	accepted.setblocking(False)

	accept_address = accepted.getsockname()

	return server, accepted, client

def control_close(lac):
	# Close the listen and the connect. Accepted
	# will be closed by SocketSelect.
	lac[2].close()
	lac[0].close()

#
#
BUMP = b'X'

class SocketChannel(object):
	def __init__(self, pending=None, client=None):
		'''
		This is the per-object client end of the control
		channel into the network I/O loop.
		'''
		self.pending = pending
		self.client = client

	def send(self, message, address):
		self.pending.put((message, address))
		buffered = self.client.send(BUMP)
		if buffered != 1:
			raise RuntimeError('Control channel not accepting commands.')

# Damn. Sent from sockets thread to creator. They
# need it to inject messages into loop.
SOCKET_CHANNEL_SCHEMA = {
	'pending': ar.Any,
	'client': ar.Any,
}

ar.bind(SocketChannel, object_schema=SOCKET_CHANNEL_SCHEMA, copy_before_sending=False)

#
#
select_table = {
	# Handling of inbound control messages.
	(ControlChannel, ReceiveBlock):	 ControlChannel_ReceiveBlock,		# Signals down the control channel.
	(ControlChannel, BrokenTransport):  ControlChannel_BrokenTransport,

	# Made to look as if the select thread can actually receive
	# sockets signals and application messages. Called from above.
	(ControlChannel, ListenForStream):  ControlChannel_ListenForStream,		# Process signals to sockets.
	(ControlChannel, ConnectStream):	ControlChannel_ConnectStream,
	(ControlChannel, Shutdown):		 ControlChannel_Shutdown,
	(ControlChannel, Bump):			 ControlChannel_Bump,
	(ControlChannel, StopListening):	ControlChannel_StopListening,
	(ControlChannel, ar.Stop):		  ControlChannel_Stop,

	# Operational sockets
	(TcpServer,	ReceiveBlock):	   TcpServer_ReceiveBlock,			# Accept inbound connections.
	(TcpServer,	BrokenTransport):	TcpServer_BrokenTransport,

	(TcpClient,   ReceiveBlock):		TcpClient_ReceiveBlock,			# Deferred connections.
	(TcpClient,   ReadyToSend):		 TcpClient_ReadyToSend,
	(TcpClient,   BrokenTransport):	 TcpClient_BrokenTransport,

	(TcpStream,   ReceiveBlock):		TcpStream_ReceiveBlock,
	(TcpStream,   ReadyToSend):		 TcpStream_ReadyToSend,
	(TcpStream,   BrokenTransport):	 TcpStream_BrokenTransport,
}

class SocketSelect(ar.Threaded, ar.Stateless):
	def __init__(self):
		ar.Threaded.__init__(self)
		ar.Stateless.__init__(self)

		# Construct the control channel and access object.
		self.pending = sq.Queue()
		self.lac = control_channel()
		self.channel = SocketChannel(self.pending, self.lac[2])

		# Load control details into socket tables.
		self.listening = self.lac[0]
		self.accepted = self.lac[1]
		self.networking = {
			self.accepted: ControlChannel(self.accepted),	# Receives 1-byte BUMPs.
		}

		# Active socket lists for select.
		self.receiving = [self.accepted]
		self.sending = []
		self.faulting = self.receiving + self.sending

		# Live.
		self.running = True

	def clear(self, s, expected=None):
		# Remove the specified socket from operations.
		try:
			t = self.networking[s]
		except KeyError:
			self.warning('Attempt to remove unknown socket')
			return None

		if expected and not isinstance(t, expected):
			self.warning('Unexpected networking object "%s" (expecting "%s")' % (t.__class__.__name__, expected.__name__))
			return None

		del self.networking[s]
		try:
			self.receiving.remove(s)
		except ValueError:
			pass
		try:
			self.sending.remove(s)
		except ValueError:
			pass
		try:
			self.faulting.remove(s)
		except ValueError:
			pass
		s.close()
		return t

def SocketSelect_Start(self, message):
	# Provide channel details to parent for access
	# by application.
	self.send(self.channel, self.parent_address)

	def clean_sockets():
		self.receiving = [s for s in self.receiving if s.fileno() > -1]
		self.sending = [s for s in self.receiving if s.fileno() > -1]
		self.faulting = [s for s in self.receiving if s.fileno() > -1]

	while self.running or len(self.networking) > 1:
		R, S, F = select.select(self.receiving, self.sending, self.faulting)

		# IF FAILS
		# for every socket in every list --- select([sock],[],[],0)
		# to find which one has failed.

		for r in R:
			try:
				#h = id(r)
				#if h in self.handshaking:
				#	r.do_handshake()
				#	self.handshaking.discard(h)
				#	continue

				a = self.networking[r]
				c = a.__class__
				j = select_table[(c, ReceiveBlock)]
			except KeyError:
				continue
			except ValueError:
				#clean_sockets()
				continue
			j(self, a, r)

		for s in S:
			try:
				#h = id(s)
				#if h in self.handshaking:
				#	s.do_handshake()
				#	self.handshaking.discard(h)
				#	continue

				a = self.networking[s]
				c = a.__class__
				j = select_table[(c, ReadyToSend)]
			except KeyError:
				continue
			except ValueError:
				#clean_sockets()
				continue
			j(self, a, s)

		for f in F:
			try:
				#h = id(f)
				#if h in self.handshaking:
				#	f.do_handshake()
				#	self.handshaking.discard(h)
				#	continue

				a = self.networking[f]
				c = a.__class__
				j = select_table[(c, BrokenTransport)]
			except KeyError:
				continue
			except ValueError:
				#clean_sockets()
				continue
			j(self, a, f)

	control_close(self.lac)
	self.complete()

ar.bind(SocketSelect, (ar.Start,))



# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2024 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
'''.

.
'''
__docformat__ = 'restructuredtext'

import ansar.create as ar

__all__ = [
	'overlay',
	'mirror',
]

def overlay(self, source, target, blacklist=None):
	if not ar.is_message(source) or not ar.is_message(target):
		s = ar.tof(source)
		t = ar.tof(target)
		self.warning(f'Cannot overlay "{s}" on "{t}" - something not a bound type.')
		return
	s = source.__art__.value
	t = target.__art__.value
	for k, v in s.items():
		if k not in t:
			continue
		if blacklist and k in blacklist:
			continue
		a = getattr(target, k, None)
		if a is None:
			a = getattr(source, k, None)
			setattr(target, k, a)

def mirror(self, source, target, blacklist=None):
	if not ar.is_message(source) or not ar.is_message(target):
		s = ar.tof(source)
		t = ar.tof(target)
		self.warning(f'Cannot mirror "{s}" on "{t}" - something not a bound type.')
		return
	s = source.__art__.value
	t = target.__art__.value
	for k, v in s.items():
		if k not in t:
			continue
		if blacklist and k in blacklist:
			continue
		a = getattr(source, k, None)
		setattr(target, k, a)



# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

""".

.
"""
__docformat__ = 'restructuredtext'

import ansar.connect as ar

__all__ = [
	'InstanceOfProduct',
	'ProductAccess',
	'ProductLookup',
	'YourProduct',
]

InstanceOfProduct = ar.Enumeration(DEVELOPMENT=1, TESTING=2, INTEGRATION=3,
	QA=4, STAGING=5, PRODUCTION=6,
	DEMONSTRATION=7, TRAINING=8, SALES=9,
	EVALUATION=10, OTHER=11)

#
#
class ProductAccess(object):
	def __init__(self, access_ipp=None, encrypted=False, product_name=None, product_instance=None):
		self.access_ipp = access_ipp or ar.HostPort()
		self.encrypted = encrypted
		self.product_name = product_name
		self.product_instance = product_instance

class ProductLookup(object):
	def __init__(self, product_name=None, product_instance=None):
		self.product_name = product_name
		self.product_instance = product_instance

class YourProduct(object):
	def __init__(self, address=None):
		self.address = address

INSTANT_SCHEMA = {
	"access_ipp": ar.UserDefined(ar.HostPort),
	"encrypted": ar.Boolean(),
	"product_name": ar.Unicode(),
	"product_instance": InstanceOfProduct,
	"address": ar.Address(),
}

ar.bind(ProductAccess, object_schema=INSTANT_SCHEMA)
ar.bind(ProductLookup, object_schema=INSTANT_SCHEMA)
ar.bind(YourProduct, object_schema=INSTANT_SCHEMA)



# Author: Scott Woods <scott.18.ansar@gmail.com.com>
# MIT License
#
# Copyright (c) 2017-2023 Scott Woods
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
__docformat__ = 'restructuredtext'

import ansar.create as ar
from .socketry import *

__all__ = [
	'ScopeOfService',
	'Published',
	'NotPublished',
	'Subscribed',
	'Available',
	'NotAvailable',
	'Delivered',
	'NotDelivered',
	'Clear',
	'Cleared',
	'Dropped',
	'NetworkEnquiry',
	'NetworkConnect',
	'DirectoryRoute',
	'DirectoryScope',
	'DirectoryAncestry',
]

# Build the local published/subscribed objects.
#
ScopeOfService = ar.Enumeration(PROCESS=1, GROUP=2, HOST=3, LAN=4, WAN=5)

#
#
class Published(object):
	"""Session notification, server presence established.

	:param requested_name: name the server is known by
	:type requested_name: str
	:param requested_scope: highest level that name has been registered 
	:type requested_scope: enumeration
	:param listening_ipp: network presence created on behalf
	:type listening_ipp: HostPort
	:param published_at: moment of publication
	:type published_at: datetime
	"""
	def __init__(self, requested_name=None, requested_scope=ScopeOfService.WAN, listening_ipp=None, published_at=None):
		self.requested_name = requested_name
		self.requested_scope = requested_scope
		self.listening_ipp = listening_ipp or HostPort()
		self.published_at = published_at

class NotPublished(ar.Faulted):
	"""Session notification, server presence not established.

	:param requested_name: intended name of the server
	:type requested_name: str
	:param reason: short description
	:type reason: str
	"""
	def __init__(self, requested_name=None, reason=None):
		ar.Faulted.__init__(self, f'cannot publish "{requested_name}"', reason)
		self.requested_name = requested_name
		self.reason = reason

class Subscribed(object):
	"""Session notification, client search established.

	:param requested_search: name the client is looking for (regular expression)
	:type requested_search: str
	:param requested_scope: highest level that search has been registered 
	:type requested_scope: enumeration
	:param subscribed_at: moment of subscription
	:type subscribed_at: datetime
	"""
	def __init__(self, requested_search=None, requested_scope=ScopeOfService.WAN, subscribed_at=None):
		self.requested_search = requested_search
		self.requested_scope = requested_scope
		self.subscribed_at = subscribed_at

#
#
class Available(object):
	"""Session notification, transport to server established.

	:param matched_search: name the client was looking for (regular expression)
	:type matched_search: str
	:param matched_name: name the server is known by
	:type matched_name: str
	:param matched_scope: level at which match was made
	:type matched_scope: enumeration
	:param opened_at: start of messaging
	:type opened_at: datetime
	:param route_key: unique id for this session
	:type route_key: str
	:param agent_address: internal pub-sub controller
	:type agent_address: async address
	"""
	def __init__(self, matched_search=None, matched_name=None, matched_scope=None, opened_at=None, route_key=None, agent_address=None):
		self.matched_search = matched_search
		self.matched_name = matched_name
		self.matched_scope = matched_scope
		self.opened_at = opened_at
		self.route_key = route_key
		self.agent_address = agent_address

class NotAvailable(ar.Faulted):
	"""Session notification, transport to server not established.

	:param matched_search: name the client was looking for (regular expression)
	:type matched_search: str
	:param matched_name: name the server is known by
	:type matched_name: str
	:param matched_scope: level at which match was made
	:type matched_scope: enumeration
	:param route_key: unique id for this session
	:type route_key: str
	:param reason: why the session could not proceed
	:type reason: str
	:param agent_address: internal pub-sub controller
	:type agent_address: async address
	"""
	def __init__(self, matched_search=None, matched_name=None, matched_scope=None, route_key=None, reason=None, agent_address=None):
		ar.Faulted.__init__(self, 'no subsciber peer', reason)
		self.matched_search = matched_search
		self.matched_name = matched_name
		self.matched_scope = matched_scope
		self.route_key = route_key
		self.reason = reason
		self.agent_address = agent_address

class Delivered(object):
	"""Session notification, transport to client established.

	:param matched_search: name the client was looking for (regular expression)
	:type matched_search: str
	:param matched_name: name the server is known by
	:type matched_name: str
	:param matched_scope: level at which match was made
	:type matched_scope: enumeration
	:param opened_at: start of messaging
	:type opened_at: datetime
	:param route_key: unique id for this session
	:type route_key: str
	:param agent_address: internal pub-sub controller
	:type agent_address: async address
	"""
	def __init__(self, matched_search=None, matched_name=None, matched_scope=None, opened_at=None, route_key=None, agent_address=None):
		self.matched_search = matched_search
		self.matched_name = matched_name
		self.matched_scope = matched_scope
		self.opened_at = opened_at
		self.route_key = route_key
		self.agent_address = agent_address

class NotDelivered(ar.Faulted):
	"""Session notification, transport to client not established.

	:param matched_search: name the client was looking for (regular expression)
	:type matched_search: str
	:param matched_name: name the server is known by
	:type matched_name: str
	:param matched_scope: level at which match was made
	:type matched_scope: enumeration
	:param route_key: unique id for this session
	:type route_key: str
	:param reason: why the session could not proceed
	:type reason: str
	:param agent_address: internal pub-sub controller
	:type agent_address: async address
	"""
	def __init__(self, matched_search=None, matched_name=None, matched_scope=None, route_key=None, reason=None, agent_address=None):
		ar.Faulted.__init__(self, 'no publisher peer', reason)
		self.matched_search = matched_search
		self.matched_name = matched_name
		self.matched_scope = matched_scope
		self.route_key = route_key
		self.reason = reason
		self.agent_address = agent_address

class Clear(object):
	def __init__(self, session=None, value=None):
		self.session = session
		self.value = value

class Cleared(object):
	"""Session notification, shutdown of transport by local end.

	:param matched_search: name the client was looking for (regular expression)
	:type matched_search: str
	:param matched_name: name the server is known by
	:type matched_name: str
	:param matched_scope: level at which match was made
	:type matched_scope: enumeration
	:param route_key: unique id for this session
	:type route_key: str
	:param reason: why the session could not proceed
	:type reason: str
	:param value: result of session
	:type value: any
	"""
	def __init__(self, matched_search=None, matched_name=None, matched_scope=None, route_key=None, reason=None, value=None):
		ar.Faulted.__init__(self, 'peer cleared', reason)
		self.matched_search = matched_search
		self.matched_name = matched_name
		self.matched_scope = matched_scope
		self.route_key = route_key
		self.reason = reason
		self.value = value

class Dropped(ar.Faulted):
	"""Session notification, shutdown of transport by remote end.

	:param matched_search: name the client was looking for (regular expression)
	:type matched_search: str
	:param matched_name: name the server is known by
	:type matched_name: str
	:param matched_scope: level at which match was made
	:type matched_scope: enumeration
	:param route_key: unique id for this session
	:type route_key: str
	:param reason: why the session could not proceed
	:type reason: str
	"""
	def __init__(self, matched_search=None, matched_name=None, matched_scope=None, route_key=None, reason=None):
		ar.Faulted.__init__(self, 'peer dropped', reason)
		self.matched_search = matched_search
		self.matched_name = matched_name
		self.matched_scope = matched_scope
		self.route_key = route_key
		self.reason = reason

ENDING_SCHEMA = {
	'matched_search': ar.Unicode(),
	'matched_name': ar.Unicode(),
	'matched_scope': ScopeOfService,
	'opened_at': ar.WorldTime(),
	'route_key': ar.Unicode(),
	'session': ar.Any,
	'value': ar.Any,
	'reason': ar.Unicode(),
	'condition': ar.Unicode(),
	'explanation': ar.Unicode(),
	'error_text': ar.Unicode(),
	'error_code': ar.Integer8(),
	'exit_code': ar.Integer8(),
}

ar.bind(Clear, object_schema=ENDING_SCHEMA, copy_before_sending=False)
ar.bind(Cleared, object_schema=ENDING_SCHEMA, copy_before_sending=False)
ar.bind(Dropped, object_schema=ENDING_SCHEMA, copy_before_sending=False)

SHARED_SCHEMA = {
	'requested_search': ar.Unicode(),
	'requested_name': ar.Unicode(),
	'requested_scope': ScopeOfService,
	'listening_ipp': ar.UserDefined(HostPort),
	'agent_address': ar.Address(),
	'published_at': ar.WorldTime(),
	'subscribed_at': ar.WorldTime(),
}

SHARED_SCHEMA.update(ENDING_SCHEMA)

ar.bind(Subscribed, object_schema=SHARED_SCHEMA)
ar.bind(Published, object_schema=SHARED_SCHEMA)
ar.bind(Available, object_schema=SHARED_SCHEMA)
ar.bind(NotAvailable, object_schema=SHARED_SCHEMA)
ar.bind(Delivered, object_schema=SHARED_SCHEMA)
ar.bind(NotDelivered, object_schema=SHARED_SCHEMA)

# matched[route_key] = a (address of ServiceRoute)
# and
# route_key added to each find and listing set
# self.directory[service_name] = [message, set()]
# self.find[k] = [message, set(), dfa]
# where k =
# f'{subscriber_address}/{requested_search}'
class DirectoryRoute(object):
	def __init__(self, search_or_listing=None, agent_address=None, route_key=None):
		self.search_or_listing = search_or_listing
		self.agent_address = agent_address
		self.route_key = route_key or ar.default_set()

DIRECTORY_ROUTE_SCHEMA = {
	'search_or_listing': ar.Unicode(),
	'agent_address': ar.Address(),
	'route_key': ar.SetOf(ar.Unicode()),
}

ar.bind(DirectoryRoute, object_schema=DIRECTORY_ROUTE_SCHEMA)

class DirectoryScope(object):
	def __init__(self, scope=None, connect_above=None, started=None, connected=None, not_connected=None,
			listing=None, find=None, accepted=None):
		self.scope = scope
		self.connect_above = connect_above or HostPort()
		self.started = started
		self.connected = connected
		self.not_connected = not_connected
		self.listing = listing or ar.default_vector()
		self.find = find or ar.default_vector()
		self.accepted = accepted or ar.default_vector()

DIRECTORY_SCOPE_SCHEMA = {
	'scope' : ScopeOfService,
	'connect_above': ar.Any(),
	'started': ar.WorldTime(),
	'connected': ar.WorldTime(),
	'not_connected': ar.Unicode(),
	'listing': ar.VectorOf(ar.UserDefined(DirectoryRoute)),
	'find': ar.VectorOf(ar.UserDefined(DirectoryRoute)),
	'accepted': ar.VectorOf(ar.UserDefined(HostPort)),
}

ar.bind(DirectoryScope, object_schema=DIRECTORY_SCOPE_SCHEMA)

#
#
class NetworkEnquiry(object):
	def __init__(self, lineage=None):
		self.lineage = lineage or ar.default_vector()

NETWORK_ENQUIRY_SCHEMA = {
	'lineage': ar.VectorOf(DirectoryScope),
}

ar.bind(NetworkEnquiry, object_schema=NETWORK_ENQUIRY_SCHEMA)

#
#
class NetworkConnect(object):
	def __init__(self, scope=None, connect_above=None):
		self.scope = scope
		self.connect_above = connect_above

NETWORK_CONNECT_SCHEMA = {
	'scope': ScopeOfService,
	'connect_above': ar.Any(),
}

ar.bind(NetworkConnect, object_schema=NETWORK_CONNECT_SCHEMA)

#
#
class DirectoryAncestry(object):
	def __init__(self, lineage=None):
		self.lineage = lineage or ar.default_vector()

DIRECTORY_ANCESTRY_SCHEMA = {
	'lineage': ar.VectorOf(DirectoryScope),
}

ar.bind(DirectoryAncestry, object_schema=DIRECTORY_ANCESTRY_SCHEMA)
